#! /usr/bin/env python
import argparse

import pandas as pd

from . import process_functions


def retract_samples(syn, database_synid, col, remove_values):
    """
    Helper retraction function that help remove values from a column in
    a database

    params:
        syn: synapse object
        databse_synid:  synapse id of database
        col: column in database
        remove_values: list of values to remove from the column
    """
    # schema = syn.get(database_synid)
    remove_values_query = "','".join(remove_values)
    remove_rows = syn.tableQuery(
        "select %s from %s where %s in ('%s')"
        % (col, database_synid, col, remove_values_query)
    )
    if len(remove_rows.asRowSet()["rows"]) > 0:
        syn.delete(remove_rows.asRowSet())
    else:
        print("Nothing to retract")


def retract(syn, project_id):
    """
    Main retraction function

    params:
        syn: synapse object
        project_id: Synapse Project ID with a database mapping table
    """

    patientRetract = syn.tableQuery(
        "select * from %s"
        % process_functions.getDatabaseSynId(
            syn, tableName="patientRetraction", project_id=project_id
        )
    )
    patientRetractIds = patientRetract.asDataFrame()
    # grab all clinical samples that belong to patients in the patient clinical file and append to sample list
    sampleClinical = syn.tableQuery(
        "select * from %s"
        % process_functions.getDatabaseSynId(
            syn, tableName="sample", project_id=project_id
        )
    )
    sampleClinicalDf = sampleClinical.asDataFrame()
    appendSamples = sampleClinicalDf["SAMPLE_ID"][
        sampleClinicalDf["PATIENT_ID"].isin(patientRetractIds.geniePatientId)
    ]

    sampleRetract = syn.tableQuery(
        "select * from %s"
        % process_functions.getDatabaseSynId(
            syn, tableName="sampleRetraction", project_id=project_id
        )
    )
    sampleRetractIds = sampleRetract.asDataFrame()

    allRetractedSamples = pd.concat([sampleRetractIds["genieSampleId"], appendSamples])

    # Only need to retract clinical data, because the rest of the data is filtered by clinical data
    # Sample Clinical Data
    retract_samples(
        syn,
        process_functions.getDatabaseSynId(
            syn, tableName="sample", project_id=project_id
        ),
        "SAMPLE_ID",
        allRetractedSamples,
    )
    # Patient Clinical Data
    retract_samples(
        syn,
        process_functions.getDatabaseSynId(
            syn, tableName="patient", project_id=project_id
        ),
        "PATIENT_ID",
        patientRetractIds["geniePatientId"],
    )


def main():
    """
    Main block with argparse and calls the main retract function
    """
    parser = argparse.ArgumentParser(description="Sample retraction")
    parser.add_argument("--pemFile", type=str, help="Path to PEM file (genie.pem)")
    parser.add_argument("--project_id", type=str, help="Synapse Project ID to use.")
    parser.add_argument("--debug", action="store_true", help="Synapse Debug Feature")
    args = parser.parse_args()
    syn = process_functions.synLogin(args.pemFile, debug=args.debug)
    retract(syn, args.project_id)


if __name__ == "__main__":
    main()



from unittest import mock
from unittest.mock import patch

import pandas as pd
import pytest
import synapseclient

import genie_registry.maf
from genie_registry.maf import maf
from genie_registry.mafSP import mafSP

syn = mock.create_autospec(synapseclient.Synapse)

maf_class = maf(syn, "SAGE")
mafsp_class = mafSP(syn, "SAGE")


def test_invalidname_validateFilename():
    with pytest.raises(AssertionError):
        maf_class.validateFilename(["foo"])


def test_valid_validateFilename():
    assert maf_class.validateFilename(["data_mutations_extended_SAGE.txt"]) == "maf"


def test_perfect_validation():
    mafDf = pd.DataFrame(
        dict(
            CHROMOSOME=[1, 2, 3, 4, 5],
            START_POSITION=[1, 2, 3, 4, 2],
            REFERENCE_ALLELE=["A", "A", "A", "A", "A"],
            TUMOR_SAMPLE_BARCODE=[
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID1-1",
            ],
            T_ALT_COUNT=[1, 2, 3, 4, 3],
            T_DEPTH=[1, 2, 3, 4, 3],
            T_REF_COUNT=[1, 2, 3, 4, 3],
            N_DEPTH=[1, 2, 3, float("nan"), 3],
            N_REF_COUNT=[1, 2, 3, 4, 3],
            N_ALT_COUNT=[1, 2, 3, 4, 3],
            TUMOR_SEQ_ALLELE2=["A", "A", "A", "A", "A"],
        )
    )

    error, warning = maf_class._validate(mafDf)
    assert error == ""
    assert warning == ""
    error, warning = mafsp_class._validate(mafDf)
    assert error == ""
    assert warning == ""


def test_firstcolumn_validation():
    """Tests if first column isn't correct"""
    mafDf = pd.DataFrame(
        {
            "REFERENCE_ALLELE": ["A", "B", "C", "D", "E"],
            "START_POSITION": [1, 2, 3, 4, 2],
            "CHROMOSOME": ["A", "A", "A", "A", "A"],
            "TUMOR_SAMPLE_BARCODE": [
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID1-1",
                "GENIE-SAGE-ID1-1",
            ],
            "T_ALT_COUNT": [1, 2, 3, 4, 3],
            "T_DEPTH": [1, 2, 3, 4, 3],
            "T_REF_COUNT": [1, 2, 3, 4, 3],
            "N_DEPTH": [1, 2, 3, 4, 3],
            "N_REF_COUNT": [1, 2, 3, 4, 3],
            "N_ALT_COUNT": [1, 2, 3, 4, 3],
            "TUMOR_SEQ_ALLELE2": ["A", "A", "A", "A", "A"],
        }
    )
    order = [
        "REFERENCE_ALLELE",
        "START_POSITION",
        "CHROMOSOME",
        "TUMOR_SAMPLE_BARCODE",
        "T_ALT_COUNT",
        "T_DEPTH",
        "T_REF_COUNT",
        "N_DEPTH",
        "N_REF_COUNT",
        "N_ALT_COUNT",
        "TUMOR_SEQ_ALLELE2",
    ]
    error, warning = maf_class._validate(mafDf[order])
    expectedErrors = (
        "maf: First column header must be "
        "one of these: CHROMOSOME, HUGO_SYMBOL, "
        "TUMOR_SAMPLE_BARCODE.\n"
    )
    assert error == expectedErrors
    assert warning == ""


def test_missingcols_validation():
    """Tests missing columns"""
    emptydf = pd.DataFrame()
    error, warning = maf_class._validate(emptydf)
    expected_errors = (
        "maf: Must at least have these headers: "
        "CHROMOSOME,START_POSITION,REFERENCE_ALLELE,"
        "TUMOR_SAMPLE_BARCODE,T_ALT_COUNT,"
        "TUMOR_SEQ_ALLELE2. "
        "If you are writing your maf file with R, please make"
        "sure to specify the 'quote=FALSE' parameter.\n"
        "maf: If missing T_DEPTH, must have T_REF_COUNT!\n"
    )
    expected_warnings = (
        "maf: Does not have the column headers "
        "that can give extra information to the processed "
        "maf: T_REF_COUNT, N_DEPTH, N_REF_COUNT, "
        "N_ALT_COUNT.\n"
    )
    assert error == expected_errors
    assert warning == expected_warnings


def test_errors_validation():
    mafDf = pd.DataFrame(
        dict(
            CHROMOSOME=[1, "chr2", "WT", 4, 5],
            START_POSITION=[1, 2, 3, 4, 2],
            REFERENCE_ALLELE=["NA", float("nan"), "A", "A", "A"],
            TUMOR_SAMPLE_BARCODE=["ID1-1", "ID1-1", "ID1-1", "ID1-1", "ID1-1"],
            TUMOR_SEQ_ALLELE2=["B", "B", "B", "B", "B"],
            T_ALT_COUNT=[1, 2, 3, 4, 3],
            T_DEPTH=[1, 2, 3, 4, 3],
            N_REF_COUNT=[1, 2, 3, 4, 3],
            N_ALT_COUNT=[1, 2, 3, 4, 3],
        )
    )

    error, warning = maf_class._validate(mafDf)

    expectedErrors = (
        "maf: "
        "REFERENCE_ALLELE can't have any blank or null values.\n"
        "maf: "
        "CHROMOSOME column cannot have any values that start "
        "with 'chr' or any 'WT' values.\n"
        "maf: TUMOR_SAMPLE_BARCODE must start with GENIE-SAGE\n"
    )
    expectedWarnings = (
        "maf: "
        "Does not have the column headers that can give "
        "extra information to the processed maf: "
        "T_REF_COUNT, N_DEPTH.\n"
        "maf: "
        "REFERENCE_ALLELE column contains 'NA' values, "
        "which cannot be placeholders for blank values.  "
        "Please put in empty strings for blank values.\n"
    )

    assert error == expectedErrors
    assert warning == expectedWarnings


def test_invalid_validation():
    mafDf = pd.DataFrame(
        dict(
            CHROMOSOME=[1, 2, 3, 4, 2, 4],
            T_ALT_COUNT=[1, 2, 3, 4, 3, 4],
            START_POSITION=[1, 2, 3, 4, 2, 4],
            REFERENCE_ALLELE=["A ", "A ", "A", "A ", "A ", " A"],
            TUMOR_SAMPLE_BARCODE=["ID1-1", "ID1-1", "ID1-1", "ID1-1", "ID1-1", "ID1-1"],
            N_DEPTH=[1, 2, 3, 4, 3, 4],
            N_REF_COUNT=[1, 2, 3, "string", 3, 4],
            N_ALT_COUNT=[1, 2, 3, 4, 3, 4],
            TUMOR_SEQ_ALLELE2=["NA", float("nan"), " A", "A ", " A", " A"],
        )
    )

    with patch.object(
        genie_registry.maf, "_check_tsa1_tsa2", return_value=""
    ) as check_tsa1_tsa2:
        error, warning = maf_class._validate(mafDf)
        check_tsa1_tsa2.assert_called_once_with(mafDf)
    expectedErrors = (
        "maf: Must not have duplicated variants. "
        "Samples with duplicated variants: ID1-1\n"
        "maf: "
        "If missing T_DEPTH, must have T_REF_COUNT!\n"
        "maf: N_REF_COUNT must be a numerical column.\n"
        "maf: "
        "TUMOR_SEQ_ALLELE2 can't have any blank or null values.\n"
        "maf: TUMOR_SAMPLE_BARCODE must start with GENIE-SAGE\n"
    )
    expectedWarnings = (
        "maf: TUMOR_SEQ_ALLELE2 column contains 'NA' values, "
        "which cannot be placeholders for blank values.  "
        "Please put in empty strings for blank values.\n"
        "maf: Does not have the column headers that can give "
        "extra information to the processed maf: T_REF_COUNT.\n"
    )
    assert error == expectedErrors
    assert warning == expectedWarnings


@pytest.mark.parametrize("col", ["temp", "REFERENCE_ALLELE"])
def test_noerror__check_allele_col(col):
    """Test error and warning is an empty string if REF col isn't passed in"""
    df = pd.DataFrame(dict(REFERENCE_ALLELE=["A", "A"]))
    error, warning = genie_registry.maf._check_allele_col(df, col)
    assert error == ""
    assert warning == ""


def test_warning__check_allele_col():
    """Test warning occurs when 'NA' string is passed in"""
    df = pd.DataFrame(dict(TEMP=["NA", "A"]))
    error, warning = genie_registry.maf._check_allele_col(df, "TEMP")
    assert error == ""
    assert warning == (
        "maf: "
        "TEMP column contains 'NA' values, "
        "which cannot be placeholders for blank values.  "
        "Please put in empty strings for blank values.\n"
    )


def test_error__check_allele_col():
    """Test error occurs when blank allele is passed in"""
    df = pd.DataFrame(dict(TEMP=[float("nan"), "A"]))
    error, warning = genie_registry.maf._check_allele_col(df, "TEMP")
    assert error == ("maf: TEMP can't have any blank or null values.\n")
    assert warning == ""


def test_invalid__check_tsa1_tsa2():
    """Test the scenario in which maf file has TSA1 and TSA2 and fails"""
    df = pd.DataFrame(
        dict(
            REFERENCE_ALLELE=["A", "A", "A"],
            TUMOR_SEQ_ALLELE1=["B", "B", "B"],
            TUMOR_SEQ_ALLELE2=["C", "C", "C"],
        )
    )
    error = genie_registry.maf._check_tsa1_tsa2(df)
    assert error == (
        "maf: Contains both "
        "TUMOR_SEQ_ALLELE1 and TUMOR_SEQ_ALLELE2 columns. "
        "All values in TUMOR_SEQ_ALLELE1 must match all values in "
        "REFERENCE_ALLELE or all values in TUMOR_SEQ_ALLELE2.\n"
    )


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame(
            dict(
                REFERENCE_ALLELE=["A", "A", "A"],
                TUMOR_SEQ_ALLELE1=["C", "C", "C"],
                TUMOR_SEQ_ALLELE2=["C", "C", "C"],
            )
        ),
        pd.DataFrame(
            dict(
                REFERENCE_ALLELE=["C", "C", "C"],
                TUMOR_SEQ_ALLELE1=["C", "C", "C"],
                TUMOR_SEQ_ALLELE2=["A", "A", "A"],
            )
        ),
    ],
)
def test_valid__check_tsa1_tsa2(df):
    """Test valid TSA1 and TSA2"""
    error = genie_registry.maf._check_tsa1_tsa2(df)
    assert error == ""



from unittest import mock

import pytest
import synapseclient


@pytest.fixture
def genie_config():
    oncotree_link = (
        "http://oncotree.mskcc.org/api/tumorTypes/tree?version=oncotree_2017_06_21"
    )
    config = {
        "vcf2maf": "syn22493903",
        "cna": "syn11600835",
        "bed": "syn11600834",
        "seg": "syn11600836",
        "fusions": "syn11600837",
        "sample": "syn11600838",
        "patient": "syn11600839",
        "patientCounts": "syn11600832",
        "mafSP": "syn11600810",
        "mutationsInCis": "syn11601206",
        "sampleRetraction": "syn11601155",
        "patientRetraction": "syn11600807",
        "vitalStatus": "syn11600808",
        "bedSP": "syn11600809",
        "validationStatus": "syn11601227",
        "errorTracker": "syn11601244",
        "centerMapping": "syn11601248",
        "processTracker": "syn11604890",
        "clinicalSP": "syn11600812",
        "main": "syn7208886",
        "dbMapping": "syn11600968",
        "md": "syn11605077",
        "consortium": "syn11605415",
        "public": "syn11605416",
        "release": "syn11607356",
        "mafinbed": "syn11600814",
        "fileview": "syn11608914",
        "cbs": "syn11600836",
        "vcf": "syn11608914",
        "maf": "syn11608914",
        "centerMaf": "syn12279903",
        "centerMafView": "syn12292501",
        "oncotreeLink": oncotree_link,
        "releaseFolder": "syn17079016",
        "assayinfo": "syn18404286",
        "logs": "syn10155804",
        "center_config": {
            "SAGE": {
                "center": "SAGE",
                "inputSynId": "syn11601335",
                "stagingSynId": "syn11601337",
                "release": True,
                "mutationInCisFilter": "ON",
            },
            "TEST": {
                "center": "TEST",
                "inputSynId": "syn11601340",
                "stagingSynId": "syn11601342",
                "release": True,
                "mutationInCisFilter": "ON",
            },
        },
        "genie_annotation_pkg": "/path/to/nexus",
        "ethnicity_mapping": "syn7434242",
        "race_mapping": "syn7434236",
        "sex_mapping": "syn7434222",
        "sampletype_mapping": "syn7434273",
    }
    return config


@pytest.fixture()
def syn():
    return mock.create_autospec(synapseclient.Synapse)



#!/usr/bin/env python3
from collections import defaultdict
import datetime
import logging
import os
import time
from typing import List

import synapseclient  # lgtm [py/import-and-import-from]
from synapseclient import Synapse
from synapseclient.core.utils import to_unix_epoch_time
import synapseutils
import pandas as pd

from . import process_functions, process_mutation, toRetract, validate

logger = logging.getLogger(__name__)


DUPLICATED_FILE_ERROR = (
    "Duplicated filename! Files should be uploaded as new versions "
    "and the entire dataset should be uploaded."
)

"""
TODO:
Could potentially get all the inforamation of the file entity right here
To avoid the syn.get rest call later which doesn't actually download the file
"""

# def rename_file(ent):
#     '''
#     Gets file from synapse and renames the file if necessary.

#     Adds the expected name as an annotation to a Synapse File object.

#     Args:
#         synid : Synapse id or entity

#     Returns:
#         entity with annotation set for path of corrected file
#     '''
#     dirpath = os.path.dirname(ent.path)
#     expectedpath = os.path.join(dirpath, ent.name)

#     ent.annotations.expectedPath = expectedpath
#     return ent


def entity_date_to_timestamp(entity_date_time):
    """Convert Synapse object date/time string (from modifiedOn or createdOn properties) to a timestamp."""

    date_and_time = entity_date_time.split(".")[0]
    date_time_obj = datetime.datetime.strptime(date_and_time, "%Y-%m-%dT%H:%M:%S")
    return to_unix_epoch_time(date_time_obj)


def get_center_input_files(syn, synid, center, process="main", downloadFile=True):
    """
    This function walks through each center's input directory
    to get a list of tuples of center files

    Args:
        syn: Synapse object
        synid: Center input folder synid
        center: Center name
        process: Process type includes, main, vcf, maf and mafSP.
                 Defaults to main such that the vcf

    Returns:
        List of entities with the correct format to pass into validation
    """
    logger.info("GETTING {center} INPUT FILES".format(center=center))
    clinical_pair_name = [
        "data_clinical_supp_sample_{center}.txt".format(center=center),
        "data_clinical_supp_patient_{center}.txt".format(center=center),
    ]

    center_files = synapseutils.walk(syn, synid)
    clinicalpair_entities = []
    prepared_center_file_list = []

    for _, _, entities in center_files:
        for name, ent_synid in entities:
            # This is to remove vcfs from being validated during main
            # processing. Often there are too many vcf files, and it is
            # not necessary for them to be run everytime.
            if name.endswith(".vcf") and process != "mutation":
                continue

            ent = syn.get(ent_synid, downloadFile=downloadFile)

            # Clinical file can come as two files.
            # The two files need to be merged together which is
            # why there is this format

            if name in clinical_pair_name:
                clinicalpair_entities.append(ent)
                continue

            prepared_center_file_list.append([ent])

    if clinicalpair_entities:
        # clinicalpair_entities = [x for x in clinicalpair]
        prepared_center_file_list.append(clinicalpair_entities)

    return prepared_center_file_list


def check_existing_file_status(validation_status_table, error_tracker_table, entities):
    """
    This function checks input files against the existing validation and error
    tracking dataframe

    Args:
        validation_status_table: Validation status Synapse Table query result
        error_tracker_table: Error tracking Synapse Table query result
        entities: list of center input entites

    Returns:
        dict: Input file status
            status_list: file validation status
            error_list: Errors of the files if they exist,
            to_validate: Boolean value for whether of not an input
                         file needs to be validated
    """
    if len(entities) > 2:
        raise ValueError("There should never be more than 2 files being validated.")

    statuses = []
    errors = []

    validation_statusdf = validation_status_table.asDataFrame()
    error_trackerdf = error_tracker_table.asDataFrame()
    # This should be outside fo the forloop so that it doesn't
    # get reset
    to_validate = False
    for ent in entities:
        # Get the current status and errors from the tables.
        current_status = validation_statusdf[validation_statusdf["id"] == ent.id]
        current_error = error_trackerdf[error_trackerdf["id"] == ent.id]

        if current_status.empty:
            to_validate = True
        else:
            # This to_validate is here, because the following is a
            # sequential check of whether files need to be validated
            statuses.append(current_status["status"].values[0])
            if current_error.empty:
                to_validate = current_status["status"].values[0] == "INVALID"
            else:
                errors.append(current_error["errors"].values[0])
            # Add Name check here (must add name of the entity as a column)
            if (
                current_status["md5"].values[0] != ent.md5
                or current_status["name"].values[0] != ent.name
            ):
                to_validate = True
            else:
                status_str = "{filename} ({id}) FILE STATUS IS: {filestatus}"
                logger.info(
                    status_str.format(
                        filename=ent.name,
                        id=ent.id,
                        filestatus=current_status["status"].values[0],
                    )
                )

    return {"status_list": statuses, "error_list": errors, "to_validate": to_validate}


def _send_validation_error_email(syn, user, message_objs):
    """
    Sends validation error email

    Args:
        syn: Synapse object
        user: username to send message to
        message_objs: list of dicts with 'filenames' and 'messages' to send
    """

    username = syn.getUserProfile(user)["userName"]

    errors = ""
    for message_obj in message_objs:
        file_names = ", ".join(message_obj["filenames"])
        error_message = message_obj["messages"]
        errors += f"Filenames: {file_names}, Errors:\n {error_message}\n\n"

    email_message = (
        f"Dear {username},\n\n"
        "You have invalid files! "
        f"Here are the reasons why:\n\n{errors}"
    )

    date_now = datetime.datetime.today().strftime("%Y-%m-%d %H:%M:%S")

    syn.sendMessage(
        userIds=[user],
        messageSubject=f"GENIE Validation Error - {date_now}",
        messageBody=email_message,
    )


def _get_status_and_error_list(valid, message, entities):
    """
    Helper function to return the status and error list of the
    files based on validation result.

    Args:
        valid: Boolean value of results of validation
        message: Validation message
        entities: List of Synapse Entities

    Returns:
        tuple: input_status_list - status of input files list,
               invalid_errors_list - error list
    """
    if valid:
        input_status_list = [{"entity": ent, "status": "VALIDATED"} for ent in entities]
        invalid_errors_list = []
    else:
        input_status_list = [{"entity": ent, "status": "INVALID"} for ent in entities]
        invalid_errors_list = [{"entity": ent, "errors": message} for ent in entities]
    return input_status_list, invalid_errors_list


def validatefile(
    syn,
    project_id,
    entities,
    validation_status_table,
    error_tracker_table,
    center,
    format_registry=None,
    genie_config=None,
):
    """Validate a list of entities.

    If a file has not changed, then it doesn't need to be validated.

    Args:
        syn: Synapse object
        entities: A list of entities for a single file 'type' (usually a single file, but clinical can have two)
        validation_statusdf: Validation status dataframe
        error_trackerdf: Invalid files error tracking dataframe
        center: Center of interest
        oncotree_link: Oncotree url

    Returns:
        tuple: input_status_list - status of input files,
               invalid_errors_list - error list
               messages_to_send - list of tuples with (filenames, message, file_users)

    """

    # filepaths = [entity.path for entity in entities]
    filenames = [entity.name for entity in entities]

    logger.info("VALIDATING {filenames}".format(filenames=", ".join(filenames)))

    file_users = [entities[0].modifiedBy, entities[0].createdBy]

    check_file_status = check_existing_file_status(
        validation_status_table, error_tracker_table, entities
    )

    status_list = check_file_status["status_list"]
    error_list = check_file_status["error_list"]

    messages_to_send = []

    # Need to figure out to how to remove this
    # This must pass in filenames, because filetype is determined by entity
    # name Not by actual path of file
    validator = validate.GenieValidationHelper(
        syn=syn,
        project_id=project_id,
        center=center,
        entitylist=entities,
        format_registry=format_registry,
        genie_config=genie_config,
    )
    filetype = validator.file_type
    if check_file_status["to_validate"]:
        valid_cls, message = validator.validate_single_file(
            oncotree_link=genie_config["oncotreeLink"], nosymbol_check=False
        )

        logger.info("VALIDATION COMPLETE")
        input_status_list, invalid_errors_list = _get_status_and_error_list(
            valid_cls.is_valid(), message, entities
        )
        # Send email the first time the file is invalid
        if invalid_errors_list:
            messages_to_send.append((filenames, message, file_users))
    else:
        input_status_list = [
            {"entity": entity, "status": status}
            for entity, status in zip(entities, status_list)
        ]
        invalid_errors_list = [
            {"entity": entity, "errors": errors}
            for entity, errors in zip(entities, error_list)
        ]
    # add in static filetype and center information
    for input_status in input_status_list:
        input_status.update({"fileType": filetype, "center": center})
    # An empty list is returned if there are no errors,
    # so nothing will be appended
    for invalid_errors in invalid_errors_list:
        invalid_errors.update({"fileType": filetype, "center": center})
    return input_status_list, invalid_errors_list, messages_to_send


# TODO: Create ProcessHelper class
def processfiles(
    syn,
    validfiles,
    center,
    path_to_genie,
    processing="main",
    format_registry=None,
    genie_config=None,
):
    """Processing validated files

    Args:
        syn: Synapse object
        validfiles: pandas dataframe containing validated files
                    has 'id', 'path', and 'fileType' column
        center: GENIE center name
        path_to_genie: Path to GENIE workdir
        center_mapping_df: Center mapping dataframe
        oncotree_link: Link to oncotree
        databaseToSynIdMappingDf: Database to synapse id mapping dataframe
        processing: Processing type. Defaults to main
    """
    logger.info(f"PROCESSING {center} FILES: {len(validfiles)}")
    center_staging_folder = os.path.join(path_to_genie, center)
    center_staging_synid = genie_config["center_config"][center]["stagingSynId"]

    if not os.path.exists(center_staging_folder):
        os.makedirs(center_staging_folder)

    if processing == "main":
        for _, row in validfiles.iterrows():
            filetype = row["fileType"]
            # filename = os.path.basename(filePath)
            newpath = os.path.join(center_staging_folder, row["name"])
            # store = True
            # Table id can be None
            tableid = genie_config.get(filetype)

            if filetype is not None and filetype != "other":
                # Example GENIE config can be found in tests/conftest.py
                processor = format_registry[filetype](
                    syn=syn, center=center, genie_config=genie_config
                )
                processor.process(
                    filePath=row["path"],
                    newPath=newpath,
                    parentId=center_staging_synid,
                    databaseSynId=tableid,
                    fileSynId=row["id"],
                )
    else:
        process_mutation.process_mutation_workflow(
            syn=syn,
            center=center,
            validfiles=validfiles,
            genie_config=genie_config,
            workdir=path_to_genie,
        )

    logger.info("ALL DATA STORED IN DATABASE")


def append_duplication_errors(duplicated_filesdf, user_message_dict):
    """Duplicated files can occur because centers can upload files with the
    same filename in different folders.  This is to append duplication
    errors to the list of errors to email

    Args:
        duplicated_filesdf: Dataframe of duplciated files
        user_message_dict: Dictionary containing list of error messages to
                           send to each user.

    Returns:
        Dictionary containing list of error messages to send to each user.
    """
    duplication_error = (
        "Duplicated filename! Files should be uploaded as new versions "
        "and the entire dataset should be uploaded."
    )
    if not duplicated_filesdf.empty:
        filenames = []
        users = []
        for entity in duplicated_filesdf["entity"]:
            users.append(entity.modifiedBy)
            users.append(entity.createdBy)
            filenames.append(entity.name)
        file_messages = dict(filenames=filenames, messages=duplication_error)
        # Must get unique set of users or there
        # will be duplicated error messages sent in the email
        for user in set(users):
            user_message_dict[user].append(file_messages)
    return user_message_dict


def get_duplicated_files(validation_statusdf):
    """
    Check for duplicated files.  There should be no duplication,
    files should be uploaded as new versions and the entire dataset
    should be uploaded everytime

    #TODO: This is a custom GENIE function

    Args:
        validation_statusdf: dataframe with 'name' and 'id' column
        duplicated_error_message: Error message for duplicated files

    Returns:
        dataframe with 'id', 'name', 'errors', 'center', 'fileType'
        and 'entity' of duplicated files
    """
    # This is special
    logger.info("CHECK FOR DUPLICATED FILES")
    duplicated_filesdf = validation_statusdf[
        validation_statusdf["name"].duplicated(keep=False)
    ]
    # Define filename str vector
    filename_str = validation_statusdf.name.str
    # cbs/seg files should not be duplicated.
    cbs_seg_index = filename_str.endswith(("cbs", "seg"))
    cbs_seg_files = validation_statusdf[cbs_seg_index]
    if len(cbs_seg_files) > 1:
        duplicated_filesdf = pd.concat([duplicated_filesdf, cbs_seg_files])
    # clinical files should not be duplicated.
    clinical_index = filename_str.startswith("data_clinical_supp")
    clinical_files = validation_statusdf[clinical_index]
    if len(clinical_files) > 2:
        duplicated_filesdf = pd.concat([duplicated_filesdf, clinical_files])
    duplicated_filesdf.drop_duplicates("id", inplace=True)
    logger.info("THERE ARE {} DUPLICATED FILES".format(len(duplicated_filesdf)))
    duplicated_filesdf["errors"] = DUPLICATED_FILE_ERROR
    return duplicated_filesdf


def build_validation_status_table(input_valid_statuses: List[dict]):
    """Build validation status dataframe

    Args:
        input_valid_statuses: list of file validation status

    Returns:
        Validation status dataframe

    """
    status_table_columns = [
        "id",
        "path",
        "md5",
        "status",
        "name",
        "modifiedOn",
        "fileType",
        "center",
        "entity",
    ]
    input_status_rows = []
    for input_status in input_valid_statuses:
        entity = input_status["entity"]
        row = {
            "id": entity.id,
            "path": entity.path,
            "md5": entity.md5,
            "status": input_status["status"],
            "name": entity.name,
            "modifiedOn": entity_date_to_timestamp(entity.properties.modifiedOn),
            "fileType": input_status["fileType"],
            "center": input_status["center"],
            "entity": entity,
        }
        input_status_rows.append(row)
    if input_status_rows:
        input_valid_statusdf = pd.DataFrame(input_status_rows)
    else:
        input_valid_statusdf = pd.DataFrame(
            input_status_rows, columns=status_table_columns
        )
    return input_valid_statusdf


def build_error_tracking_table(invalid_errors: List[dict]):
    """Build error tracking dataframe

    Args:
        invalid_errors: list of file invalid errors

    Returns:
        Error tracking dataframe

    """
    error_table_columns = ["id", "errors", "name", "fileType", "center", "entity"]
    invalid_error_rows = []
    for invalid_error in invalid_errors:
        entity = invalid_error["entity"]
        row = {
            "id": entity.id,
            "errors": invalid_error["errors"],
            "name": entity.name,
            "fileType": invalid_error["fileType"],
            "center": invalid_error["center"],
            "entity": entity,
        }
        invalid_error_rows.append(row)
    if invalid_error_rows:
        invalid_errorsdf = pd.DataFrame(invalid_error_rows)
    else:
        invalid_errorsdf = pd.DataFrame(invalid_error_rows, columns=error_table_columns)
    return invalid_errorsdf


def update_status_and_error_tables(
    syn,
    input_valid_statusdf,
    invalid_errorsdf,
    validation_status_table,
    error_tracker_table,
):
    """
    Update validation status and error tracking table

    Args:
        syn: Synapse object
        center: Center
        input_valid_status: list of lists of validation status
        invalid_errors: List of lists of invalid errors
        validation_status_table: Synapse table query of validation status
        error_tracker_table: Synapse table query of error tracker

    """
    logger.info("UPDATE VALIDATION STATUS DATABASE")

    process_functions.updateDatabase(
        syn,
        error_tracker_table.asDataFrame(),
        invalid_errorsdf,
        error_tracker_table.tableId,
        ["id"],
        to_delete=True,
    )

    process_functions.updateDatabase(
        syn,
        validation_status_table.asDataFrame(),
        input_valid_statusdf,
        validation_status_table.tableId,
        ["id"],
        to_delete=True,
    )


def _update_tables_content(validation_statusdf, error_trackingdf):
    """Update validation status and error tracking dataframes with duplicated
    files.  Also update the error table to only contain errors - centers
    may have fixed their files so will want to remove old errors.

    Args:
        validation_statusdf: Validation status dataframe
        error_trackingdf: Error tracking dataframe

    Returns:
        dict: validation_statusdf: Updated validation status dataframe
              error_trackingdf: Updated error tracking dataframe
              duplicated_filesdf:  Duplicated files dataframe

    """
    # Get duplicated files
    duplicated_filesdf = get_duplicated_files(validation_statusdf)
    # index of all duplicated files
    duplicated_idx = validation_statusdf["id"].isin(duplicated_filesdf["id"])
    validation_statusdf["status"][duplicated_idx] = "INVALID"
    duplicated_idx = error_trackingdf["id"].isin(duplicated_filesdf["id"])
    error_trackingdf["errors"][duplicated_idx] = DUPLICATED_FILE_ERROR

    # Old errors are pulled down in validation, so obtain list of
    # files with duplicated file errors
    dup_ids = error_trackingdf["id"][
        error_trackingdf["errors"] == DUPLICATED_FILE_ERROR
    ]
    # Checks to see if the old duplicated files are still duplicated
    remove_ids = dup_ids[~dup_ids.isin(duplicated_filesdf["id"])]

    # Remove fixed duplicated files
    error_trackingdf = error_trackingdf[~error_trackingdf["id"].isin(remove_ids)]
    validation_statusdf = validation_statusdf[
        ~validation_statusdf["id"].isin(remove_ids)
    ]

    # Append duplicated file errors
    duplicated_filesdf["id"].isin(error_trackingdf["id"][duplicated_idx])
    error_trackingdf = pd.concat(
        [error_trackingdf, duplicated_filesdf[error_trackingdf.columns]]
    )
    # Remove duplicates if theres already an error that exists for the file
    error_trackingdf.drop_duplicates("id", inplace=True)

    # Since old errors are retained, make sure to only update
    # files that are actually invalid
    invalid_ids = validation_statusdf["id"][validation_statusdf["status"] == "INVALID"]
    error_trackingdf = error_trackingdf[error_trackingdf["id"].isin(invalid_ids)]
    # Fill blank file type values with 'other'
    error_trackingdf["fileType"].fillna("other", inplace=True)
    validation_statusdf["fileType"].fillna("other", inplace=True)

    return {
        "validation_statusdf": validation_statusdf,
        "error_trackingdf": error_trackingdf,
        "duplicated_filesdf": duplicated_filesdf,
    }


def validation(
    syn, project_id, center, process, center_files, format_registry, genie_config
) -> pd.DataFrame:
    """
    Validation of all center files

    Args:
        syn: Synapse object
        center: Center name
        process: main, mutation
        center_files:
        format_registry:
        genie_config:

    Returns:
        pd.DataFrame: Dataframe of valid GENIE files
    """
    logger.info(f"{center} has uploaded {len(center_files)} files.")
    validation_status_synid = genie_config["validationStatus"]
    error_tracker_synid = genie_config["errorTracker"]

    # Make sure the vcf validation statuses don't get wiped away
    # If process is not vcf, the vcf files are not downloaded
    # TODO: Add parameter to exclude types
    exclude_type = "vcf" if process != "mutation" else ""
    # id, md5, status, name, center, modifiedOn, fileType
    validation_status_table = syn.tableQuery(
        f"SELECT * FROM {validation_status_synid} where "
        f"center = '{center}' and fileType <> '{exclude_type}'"
    )
    # id, center, errors, name, fileType
    error_tracker_table = syn.tableQuery(
        f"SELECT * FROM {error_tracker_synid} where "
        f"center = '{center}' and fileType <> '{exclude_type}'"
    )

    input_valid_statuses = []
    invalid_errors = []

    # This default dict will capture all the error messages to send to
    # particular users
    user_message_dict = defaultdict(list)

    for ents in center_files:
        status, errors, messages_to_send = validatefile(
            syn=syn,
            project_id=project_id,
            entities=ents,
            validation_status_table=validation_status_table,
            error_tracker_table=error_tracker_table,
            center=center,
            format_registry=format_registry,
            genie_config=genie_config,
        )

        input_valid_statuses.extend(status)
        if errors is not None:
            invalid_errors.extend(errors)

        if messages_to_send:
            logger.debug("Collating messages to send to users.")
            for filenames, messages, users in messages_to_send:
                file_messages = dict(filenames=filenames, messages=messages)
                # Must get unique set of users or there
                # will be duplicated error messages sent in the email
                for user in set(users):
                    user_message_dict[user].append(file_messages)

    validation_statusdf = build_validation_status_table(input_valid_statuses)
    error_trackingdf = build_error_tracking_table(invalid_errors)

    new_tables = _update_tables_content(validation_statusdf, error_trackingdf)
    validation_statusdf = new_tables["validation_statusdf"]
    error_trackingdf = new_tables["error_trackingdf"]
    duplicated_filesdf = new_tables["duplicated_filesdf"]

    # In GENIE, we not only want to send out file format errors, but
    # also when there are duplicated errors.  The function below will
    # append duplication errors as an email to send to users (if applicable)
    user_message_dict = append_duplication_errors(duplicated_filesdf, user_message_dict)

    for user, message_objs in user_message_dict.items():
        logger.debug("Sending messages to user {user}.".format(user=user))

        _send_validation_error_email(syn=syn, user=user, message_objs=message_objs)
    # \n write out new lines when they exist in the middle of a column
    # So the \n never gets uploaded into synapse table
    # change the delimiting to '|'.
    error_trackingdf["errors"] = [
        error.replace("\n", "|") for error in error_trackingdf["errors"]
    ]
    update_status_and_error_tables(
        syn=syn,
        input_valid_statusdf=validation_statusdf,
        invalid_errorsdf=error_trackingdf,
        validation_status_table=validation_status_table,
        error_tracker_table=error_tracker_table,
    )

    valid_filesdf = validation_statusdf.query('status == "VALIDATED"')
    return valid_filesdf[["id", "path", "fileType", "name"]]


# TODO: probably should rename this for clarity
def center_input_to_database(
    syn: Synapse,
    project_id: str,
    center: str,
    process: str,
    only_validate: bool,
    delete_old: bool = False,
    format_registry: list = None,
    genie_config: dict = None,
):
    """Processing per center

    Args:
        syn (Synapse): Synapse connection
        project_id (str): GENIE Synapse project id
        center (str): GENIE center
        process (str): main or mutation processing
        only_validate (bool): Only validate or not
        delete_old (bool, optional): Delete old files. Defaults to False.
        format_registry (typing.List, optional): GENIE file format registry.
                                                 Defaults to None.
        genie_config (typing.Dict, optional): See example of genie config at
                                              ./genie_config.json. Defaults to None.
    """

    if only_validate:
        log_path = os.path.join(
            process_functions.SCRIPT_DIR, f"{center}_validation_log.txt"
        )
    else:
        log_path = os.path.join(
            process_functions.SCRIPT_DIR, f"{center}_{process}_log.txt"
        )
    # Set up logger to write to a log file as well as streaming logs
    logFormatter = logging.Formatter(
        "%(asctime)s [%(name)s][%(levelname)s] %(message)s"
    )
    fileHandler = logging.FileHandler(log_path, mode="w")
    fileHandler.setFormatter(logFormatter)
    logger.addHandler(fileHandler)

    # ----------------------------------------
    # Start processing
    # ----------------------------------------

    # path_to_genie = os.path.realpath(
    #   os.path.join(process_functions.SCRIPT_DIR, "../")
    # )
    # HACK:
    # Make the synapse cache dir the genie input folder for now
    # The main reason for this is because the .synaspeCache dir
    # is mounted by batch
    path_to_genie = os.path.expanduser("~/.synapseCache")
    # Create input and staging folders
    if not os.path.exists(os.path.join(path_to_genie, center, "input")):
        os.makedirs(os.path.join(path_to_genie, center, "input"))
    if not os.path.exists(os.path.join(path_to_genie, center, "staging")):
        os.makedirs(os.path.join(path_to_genie, center, "staging"))

    if delete_old:
        process_functions.rmFiles(os.path.join(path_to_genie, center))

    center_input_synid = genie_config["center_config"][center]["inputSynId"]
    logger.info("Center: " + center)
    center_files = get_center_input_files(syn, center_input_synid, center, process)

    # only validate if there are center files
    if center_files:
        validFiles = validation(
            syn=syn,
            project_id=project_id,
            center=center,
            process=process,
            center_files=center_files,
            format_registry=format_registry,
            genie_config=genie_config,
        )
    else:
        logger.info(f"{center} has not uploaded any files")
        return

    if len(validFiles) > 0 and not only_validate:
        # Reorganize so BED file are always validated and processed first
        bed_files = validFiles["fileType"] == "bed"
        beds = validFiles[bed_files]
        validFiles = pd.concat([beds, validFiles])
        validFiles.drop_duplicates(inplace=True)
        # merge clinical files into one row
        clinical_ind = validFiles["fileType"] == "clinical"
        if clinical_ind.any():
            clinical_files = validFiles[clinical_ind].to_dict(orient="list")
            # The [] implies the values in the dict as a list
            merged_clinical = pd.DataFrame([clinical_files])
            merged_clinical["fileType"] = "clinical"
            merged_clinical["name"] = f"data_clinical_supp_{center}.txt"
            validFiles = pd.concat([validFiles[~clinical_ind], merged_clinical])

        processTrackerSynId = genie_config["processTracker"]
        # Add process tracker for time start
        processTrackerDf = process_functions.get_syntabledf(
            syn=syn,
            query_string=(
                f"SELECT timeStartProcessing FROM {processTrackerSynId} "
                f"where center = '{center}' and "
                f"processingType = '{process}'"
            ),
        )
        if processTrackerDf.empty:
            new_rows = [
                [
                    center,
                    str(int(time.time() * 1000)),
                    str(int(time.time() * 1000)),
                    process,
                ]
            ]

            syn.store(synapseclient.Table(processTrackerSynId, new_rows))
        else:
            processTrackerDf["timeStartProcessing"].iloc[0] = str(
                int(time.time() * 1000)
            )
            syn.store(synapseclient.Table(processTrackerSynId, processTrackerDf))

        # Start transformations
        processfiles(
            syn=syn,
            validfiles=validFiles,
            center=center,
            path_to_genie=path_to_genie,
            processing=process,
            format_registry=format_registry,
            genie_config=genie_config,
        )

        # Should add in this process end tracking before the deletion of samples
        processTrackerDf = process_functions.get_syntabledf(
            syn=syn,
            query_string=(
                f"SELECT timeEndProcessing FROM {processTrackerSynId} "
                f"where center = '{center}' and "
                f"processingType = '{process}'"
            ),
        )
        processTrackerDf["timeEndProcessing"].iloc[0] = str(int(time.time() * 1000))
        syn.store(synapseclient.Table(processTrackerSynId, processTrackerDf))

        logger.info("SAMPLE/PATIENT RETRACTION")
        toRetract.retract(syn, project_id=project_id)

    else:
        messageOut = (
            f"{center} does not have any valid files"
            if not only_validate
            else f"ONLY VALIDATION OCCURED FOR {center}"
        )
        logger.info(messageOut)

    # Store and remove log file
    syn.store(synapseclient.File(log_path, parentId=genie_config["logs"]))
    os.remove(log_path)
    logger.info("ALL PROCESSES COMPLETE")



from io import StringIO
import logging
import os

from genie.example_filetype_format import FileTypeFormat
from genie import process_functions

logger = logging.getLogger(__name__)


class StructuralVariant(FileTypeFormat):

    _fileType = "sv"

    # _validation_kwargs = ["nosymbol_check", "project_id"]

    # VALIDATE FILENAME
    def _validateFilename(self, filePath):
        assert os.path.basename(filePath[0]) == "data_sv.txt"

    def _process(self, sv_df):
        sv_df.columns = [col.upper() for col in sv_df.columns]
        # Add center column
        center = [sample_id.split("-")[1] for sample_id in sv_df["SAMPLE_ID"]]
        sv_df["CENTER"] = center
        return sv_df

    def process_steps(self, sv_df, newPath, databaseSynId):
        sv_df = self._process(sv_df)
        # TODO: test the col parameter
        process_functions.updateData(
            self.syn,
            databaseSynId,
            sv_df,
            self.center,
            toDelete=True,
            col=sv_df.columns.to_list(),
        )
        sv_df.to_csv(newPath, sep="\t", index=False)
        return newPath

    # def _validate(self, sv_df, nosymbol_check, project_id):
    def _validate(self, sv_df):
        total_error = StringIO()
        total_warning = StringIO()
        sv_df.columns = [col.upper() for col in sv_df.columns]

        have_sample_col = process_functions.checkColExist(sv_df, "SAMPLE_ID")
        if not have_sample_col:
            total_error.write("Structural Variant: Must have SAMPLE_ID column.\n")
        else:
            # if sv_df["SAMPLE_ID"].duplicated().any():
            #     total_error.write(
            #         "Structural Variant: No duplicated SAMPLE_ID allowed.\n"
            #     )
            # TODO: switch to validate_genie_identifier function
            # After GH-444 is merged
            errors = process_functions.validate_genie_identifier(
                identifiers=sv_df["SAMPLE_ID"],
                center=self.center,
                filename="Structural Variant",
                col="SAMPLE_ID",
            )
            total_error.write(errors)

        if sv_df.duplicated().any():
            total_error.write("Structural Variant: No duplicated rows allowed.\n")

        warn, error = process_functions.check_col_and_values(
            sv_df,
            "SV_STATUS",
            ["SOMATIC", "GERMLINE"],
            "Structural Variant",
            required=True,
        )
        total_warning.write(warn)
        total_error.write(error)

        have_hugo_1 = process_functions.checkColExist(sv_df, "SITE1_HUGO_SYMBOL")
        have_hugo_2 = process_functions.checkColExist(sv_df, "SITE2_HUGO_SYMBOL")
        have_entrez_1 = process_functions.checkColExist(sv_df, "SITE1_ENTREZ_GENE_ID")
        have_entrez_2 = process_functions.checkColExist(sv_df, "SITE2_ENTREZ_GENE_ID")

        if not ((have_hugo_1 or have_entrez_1) and (have_hugo_2 or have_entrez_2)):
            total_error.write(
                "Structural Variant: Either SITE1_HUGO_SYMBOL/SITE1_ENTREZ_GENE_ID "
                "or SITE2_HUGO_SYMBOL/SITE2_ENTREZ_GENE_ID is required.\n"
            )

        # optional_columns = [
        #     "SITE1_REGION_NUMBER",
        #     "SITE2_REGION_NUMBER",
        #     "SITE1_REGION",
        #     "SITE2_REGION",
        #     "SITE1_CHROMOSOME",
        #     "SITE2_CHROMOSOME",
        #     "SITE1_CONTIG",
        #     "SITE2_CONTIG",
        #     "SITE1_POSITION",
        #     "SITE2_POSITION",
        #     "SITE1_DESCRIPTION",
        #     "SITE2_DESCRIPTION",
        #     "SITE2_EFFECT_ON_FRAME",
        #     "NCBI_BUILD",
        #     "CLASS",
        #     "TUMOR_SPLIT_READ_COUNT",
        #     "TUMOR_PAIRED_END_READ_COUNT",
        #     "EVENT_INFO",
        #     "BREAKPOINT_TYPE",
        #     "CONNECTION_TYPE",
        #     "ANNOTATION",
        #     "DNA_SUPPORT",
        #     "RNA_SUPPORT",
        #     "SV_LENGTH",
        #     "NORMAL_READ_COUNT",
        #     "TUMOR_READ_COUNT",
        #     "NORMAL_VARIANT_COUNT",
        #     "TUMOR_VARIANT_COUNT",
        #     "NORMAL_PAIRED_END_READ_COUNT",
        #     "NORMAL_SPLIT_READ_COUNT",
        #     "COMMENTS",
        # ]
        # Check for columns that should be integar columsn
        int_cols = [
            "SITE1_ENTREZ_GENE_ID",
            "SITE2_ENTREZ_GENE_ID",
            "SITE1_REGION_NUMBER",
            "SITE2_REGION_NUMBER",
            "SITE1_POSITION",
            "SITE2_POSITION",
            "TUMOR_SPLIT_READ_COUNT",
            "TUMOR_PAIRED_END_READ_COUNT",
            "SV_LENGTH",
            "NORMAL_READ_COUNT",
            "TUMOR_READ_COUNT",
            "NORMAL_VARIANT_COUNT",
            "TUMOR_VARIANT_COUNT",
            "NORMAL_PAIRED_END_READ_COUNT",
            "NORMAL_SPLIT_READ_COUNT",
        ]
        # Get all columns that are non integers.
        non_ints = [
            col
            for col in int_cols
            if sv_df.get(col) is not None and sv_df[col].dtype != int
        ]
        if len(non_ints) > 0:
            total_error.write(
                "Structural Variant: Only integers allowed in these "
                "column(s): {}.\n".format(", ".join(non_ints))
            )

        region_allow_vals = ["5_PRIME_UTR", "3_PRIME_UTR", "PROMOTER", "EXON", "INTRON"]
        warn, error = process_functions.check_col_and_values(
            sv_df,
            "SITE1_REGION",
            region_allow_vals,
            "Structural Variant",
            required=False,
        )
        warn, error = process_functions.check_col_and_values(
            sv_df,
            "SITE2_REGION",
            region_allow_vals,
            "Structural Variant",
            required=False,
        )
        warn, error = process_functions.check_col_and_values(
            sv_df,
            "NCBI_BUILD",
            ["GRCh37", "GRCh38"],
            "Structural Variant",
            required=False,
        )
        # total_warning.write(warn)
        total_error.write(error)

        warn, error = process_functions.check_col_and_values(
            sv_df,
            "BREAKPOINT_TYPE",
            ["PRECISE", "IMPRECISE"],
            "Structural Variant",
            required=False,
        )
        # total_warning.write(warn)
        total_error.write(error)

        warn, error = process_functions.check_col_and_values(
            sv_df,
            "CONNECTION_TYPE",
            ["3to5", "5to3", "5to5", "3to3"],
            "Structural Variant",
            required=False,
        )
        # total_warning.write(warn)
        total_error.write(error)

        warn, error = process_functions.check_col_and_values(
            sv_df, "DNA_SUPPORT", ["yes", "no"], "Structural Variant", required=False
        )
        # total_warning.write(warn)
        total_error.write(error)

        warn, error = process_functions.check_col_and_values(
            sv_df, "RNA_SUPPORT", ["yes", "no"], "Structural Variant", required=False
        )
        # total_warning.write(warn)
        total_error.write(error)

        return total_error.getvalue(), total_warning.getvalue()



from datetime import datetime
from unittest import mock
from unittest.mock import Mock, patch

import pandas as pd
import pytest
import synapseclient
import synapseutils

from genie import (
    input_to_database,
    example_filetype_format,
    process_functions,
    process_mutation,
)
from genie_registry.clinical import Clinical
from genie.validate import GenieValidationHelper


syn = mock.create_autospec(synapseclient.Synapse)
sample_clinical_synid = "syn2222"

sample_clinical_entity = synapseclient.File(
    path="data_clinical_supp_sample_SAGE.txt",
    id=sample_clinical_synid,
    parentId="syn45678",
    name="data_clinical_supp_sample_SAGE.txt",
    modifiedOn="2019-03-24T12:00:00.Z",
    md5="44444",
)

patient_clinical_synid = "syn11111"
patient_clinical_entity = synapseclient.File(
    path="data_clinical_supp_patient_SAGE.txt",
    id=patient_clinical_synid,
    parentId="syn45678",
    name="data_clinical_supp_patient_SAGE.txt",
)

vcf1synid = "syn6666"
vcf1_entity = synapseclient.File(
    path="GENIE-SAGE-1-1.vcf",
    id=vcf1synid,
    parentId="syn45678",
    name="GENIE-SAGE-1-1.vcf",
)
vcf2synid = "syn8888"
vcf2_entity = synapseclient.File(
    path="GENIE-SAGE-2-1.vcf",
    id=vcf2synid,
    parentId="syn45678",
    name="GENIE-SAGE-2-1.vcf",
)
first = (
    [("inputs", "syn12345")],
    [("vcfs", "syn33333")],
    [
        ("data_clinical_supp_sample_SAGE.txt", sample_clinical_synid),
        ("data_clinical_supp_patient_SAGE.txt", patient_clinical_synid),
    ],
)
second = (
    [("vcfs", "syn33333")],
    [],
    [("GENIE-SAGE-000-1111.vcf", vcf1synid), ("GENIE-SAGE-111-2222.vcf", vcf2synid)],
)
center = "SAGE"
oncotree_link = (
    "http://oncotree.mskcc.org/api/tumorTypes/tree?version=oncotree_2017_06_21"
)
center_input_synid = "syn9999"
center_staging_synid = "syn9999"
center_mapping = {
    "inputSynId": [center_input_synid],
    "stagingSynId": [center_input_synid],
    "center": [center],
}
center_mapping_df = pd.DataFrame(center_mapping)
validation_statusdf = pd.DataFrame(
    {
        "id": ["syn1234", "syn2345"],
        "status": ["VALID", "INVALID"],
        "md5": ["3333", "44444"],
        "name": ["first.txt", "second.txt"],
        "fileType": ["filetype1", "filetype2"],
    }
)
error_trackerdf = pd.DataFrame(
    {"id": ["syn2345"], "errors": ["Invalid file format"], "fileType": ["filetype1"]}
)
emptydf = pd.DataFrame(columns=["id"], dtype=str)


class mock_csv_query_result(object):
    def __init__(self, df):
        self.df = df

    def asDataFrame(self):
        return self.df


# def test_samename_rename_file():
#     '''Test that the file path is not renamed.
#     '''
#     filename = synapseclient.utils.make_bogus_data_file()
#     entity = synapseclient.File(path=filename,
#                                 id='syn012345',
#                                 parentId='syn45678',
#                                 name=os.path.basename(filename))
#     expectedpath = filename
#     new_entity = input_to_database.rename_file(entity)
#     assert new_entity.annotations.expectedPath == expectedpath
#     os.remove(filename)


# def test_diffname_rename_file():
#     '''Test that the file path is renamed.
#     '''
#     filename = synapseclient.utils.make_bogus_data_file()
#     entity = synapseclient.File(path=filename,
#                                 id='syn012345',
#                                 parentId='syn45678',
#                                 name='testname')

#     expectedpath = os.path.join(os.path.dirname(filename), "testname")
#     new_entity = input_to_database.rename_file(entity)
#     assert new_entity.annotations.expectedPath == expectedpath
#     os.remove(filename)


def walk_return():
    """
    Generator returned by synapseutils.walk
    """
    yield first
    yield second


def walk_return_empty():
    """
    Generator returned by synapseutils.walk
    """
    yield ([], [], [])


def test_main_get_center_input_files():
    """
    Test to make sure center input files are gotten
    excluding the vcf files since process main is specified
    """
    syn_get_effects = [sample_clinical_entity, patient_clinical_entity]
    expected_center_file_list = [syn_get_effects]

    calls = [
        mock.call(sample_clinical_synid, downloadFile=True),
        mock.call(patient_clinical_synid, downloadFile=True),
    ]

    with patch.object(
        synapseutils, "walk", return_value=walk_return()
    ) as patch_synapseutils_walk, patch.object(
        syn, "get", side_effect=syn_get_effects
    ) as patch_syn_get:
        center_file_list = input_to_database.get_center_input_files(
            syn, "syn12345", center
        )

        assert len(center_file_list) == len(expected_center_file_list)
        assert len(center_file_list[0]) == 2
        assert center_file_list == expected_center_file_list
        patch_synapseutils_walk.assert_called_once_with(syn, "syn12345")
        patch_syn_get.assert_has_calls(calls)


def test_mutation_get_center_input_files():
    """
    Test to make sure center input files are gotten
    including the vcf files since process vcf is specified
    """
    syn_get_effects = [
        sample_clinical_entity,
        patient_clinical_entity,
        vcf1_entity,
        vcf2_entity,
    ]
    expected_center_file_list = [
        [vcf1_entity],
        [vcf2_entity],
        [sample_clinical_entity, patient_clinical_entity],
    ]
    calls = [
        mock.call(sample_clinical_synid, downloadFile=True),
        mock.call(patient_clinical_synid, downloadFile=True),
        mock.call(vcf1synid, downloadFile=True),
        mock.call(vcf2synid, downloadFile=True),
    ]

    with patch.object(
        synapseutils, "walk", return_value=walk_return()
    ) as patch_synapseutils_walk, patch.object(
        syn, "get", side_effect=syn_get_effects
    ) as patch_syn_get:
        center_file_list = input_to_database.get_center_input_files(
            syn, "syn12345", center, process="mutation"
        )
        assert len(center_file_list) == len(expected_center_file_list)
        assert len(center_file_list[2]) == 2
        assert center_file_list == expected_center_file_list
        patch_synapseutils_walk.assert_called_once_with(syn, "syn12345")
        patch_syn_get.assert_has_calls(calls)


def test_empty_get_center_input_files():
    """
    Test that center input files is empty if directory
    pass in is empty
    """
    with patch.object(
        synapseutils, "walk", return_value=walk_return_empty()
    ) as patch_synapseutils_walk:
        center_file_list = input_to_database.get_center_input_files(
            syn, "syn12345", center, process="vcf"
        )
        assert center_file_list == []
        patch_synapseutils_walk.assert_called_once_with(syn, "syn12345")


# @pytest.fixture(params=[
#     # tuple with (input, expectedOutput)
#     (["data_CNA_SAGE.txt"], "cna"),
#     (["data_clinical_supp_SAGE.txt"], "clinical"),
#     (["data_clinical_supp_sample_SAGE.txt",
#       "data_clinical_supp_patient_SAGE.txt"], "clinical")])
# def filename_fileformat_map(request):
#     return request.param


# def test_perfect_get_filetype(filename_fileformat_map):
#     (filepath_list, fileformat) = filename_fileformat_map
#     assert input_to_database.get_filetype(
#         syn, filepath_list, center) == fileformat


# def test_wrongfilename_get_filetype():
#     assert input_to_database.get_filetype(syn, ['wrong.txt'], center) is None


def test_unvalidatedinput_check_existing_file_status():
    """
    Test the values returned by input that hasn't be validated
    """
    entity = synapseclient.Entity(id="syn1234")
    entities = [entity]

    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(emptydf), mock_csv_query_result(emptydf), entities
    )
    assert file_status["to_validate"]
    assert file_status["status_list"] == []
    assert file_status["error_list"] == []


def test_valid_check_existing_file_status():
    """
    Test the values returned by input that is already valid
    """
    entity = synapseclient.Entity(name="first.txt", id="syn1234", md5="3333")
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf),
        mock_csv_query_result(error_trackerdf),
        entities,
    )
    assert not file_status["to_validate"]
    assert file_status["status_list"] == ["VALID"]
    assert file_status["error_list"] == []


def test_invalid_check_existing_file_status():
    """
    Test the values returned by input that is invalid
    """
    entity = synapseclient.Entity(name="second.txt", id="syn2345", md5="44444")
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf),
        mock_csv_query_result(error_trackerdf),
        entities,
    )
    assert not file_status["to_validate"]
    assert file_status["status_list"] == ["INVALID"]
    assert file_status["error_list"] == ["Invalid file format"]


def test_nostorederrors_check_existing_file_status():
    """
    If there is no error uploaded, must re-validate file
    """
    entity = synapseclient.Entity(name="second.txt", id="syn2345", md5="44444")
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf),
        mock_csv_query_result(emptydf),
        entities,
    )
    assert file_status["to_validate"]
    assert file_status["status_list"] == ["INVALID"]
    assert file_status["error_list"] == []


def test_diffmd5validate_check_existing_file_status():
    """
    If md5 is different from stored md5, must re-validate file
    """
    entity = synapseclient.Entity(name="first.txt", id="syn1234", md5="44444")
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf),
        mock_csv_query_result(emptydf),
        entities,
    )
    assert file_status["to_validate"]
    assert file_status["status_list"] == ["VALID"]
    assert file_status["error_list"] == []


def test_diffnametovalidate_check_existing_file_status():
    """
    If name is different from stored name, must re-validate file
    """
    entity = synapseclient.Entity(name="second.txt", id="syn1234", md5="3333")
    entities = [entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf),
        mock_csv_query_result(emptydf),
        entities,
    )
    assert file_status["to_validate"]
    assert file_status["status_list"] == ["VALID"]
    assert file_status["error_list"] == []


def test_twoinvalid_check_existing_file_status():
    """
    Test to make sure both invalid statues get passed back
    """
    validation_statusdf = pd.DataFrame(
        {
            "id": ["syn1234", "syn2345"],
            "status": ["INVALID", "INVALID"],
            "md5": ["3333", "44444"],
            "name": ["first.txt", "second.txt"],
        }
    )
    error_trackerdf = pd.DataFrame(
        {
            "id": ["syn1234", "syn2345"],
            "errors": ["Invalid file format", "Invalid formatting issues"],
        }
    )
    first_entity = synapseclient.Entity(name="first.txt", id="syn1234", md5="3333")
    second_entity = synapseclient.Entity(name="second.txt", id="syn2345", md5="44444")
    entities = [first_entity, second_entity]
    file_status = input_to_database.check_existing_file_status(
        mock_csv_query_result(validation_statusdf),
        mock_csv_query_result(error_trackerdf),
        entities,
    )
    assert not file_status["to_validate"]
    assert file_status["status_list"] == ["INVALID", "INVALID"]
    assert file_status["error_list"] == [
        "Invalid file format",
        "Invalid formatting issues",
    ]


def test_error_check_existing_file_status():
    """
    With exception of clinical files, there should never be
    more than 2 files being passed in for validation.
    """
    with pytest.raises(
        ValueError, match="There should never be more than 2 files being validated."
    ):
        entities = ["foo", "doo", "boo"]
        input_to_database.check_existing_file_status(emptydf, emptydf, entities)


def test_valid_validatefile(genie_config):
    """
    Tests the behavior of a file that gets validated that becomes
    valid
    """
    validation_statusdf = pd.DataFrame()
    error_trackerdf = pd.DataFrame()
    entity = synapseclient.Entity(
        name="data_clinical_supp_SAGE.txt",
        id="syn1234",
        md5="44444",
        path="/path/to/data_clinical_supp_SAGE.txt",
    )
    entity["modifiedOn"] = "2019-03-24T12:00:00.Z"
    # This modifiedOn translates to: 1553428800000
    entity.modifiedBy = "333"
    entity.createdBy = "444"
    entities = [entity]
    valid_results = example_filetype_format.ValidationResults(errors="", warnings="")
    message = "Is valid"
    filetype = "clinical"
    # Only a list is returned as opposed a list of lists when there are
    # invalid errors
    status_error_list_results = ([{"entity": entity, "status": "VALIDATED"}], [])
    expected_results = (
        [
            {
                "entity": entity,
                "status": "VALIDATED",
                "fileType": filetype,
                "center": center,
            }
        ],
        [],
        [],
    )
    with patch.object(
        GenieValidationHelper, "determine_filetype", return_value=filetype
    ) as patch_determine_filetype, patch.object(
        input_to_database,
        "check_existing_file_status",
        return_value={"status_list": [], "error_list": [], "to_validate": True},
    ) as patch_check, patch.object(
        GenieValidationHelper,
        "validate_single_file",
        return_value=(valid_results, message),
    ) as patch_validate, patch.object(
        input_to_database,
        "_get_status_and_error_list",
        return_value=status_error_list_results,
    ) as patch_staterror_list, patch.object(
        input_to_database, "_send_validation_error_email"
    ) as patch_send_email:

        validate_results = input_to_database.validatefile(
            syn,
            None,
            entities,
            validation_statusdf,
            error_trackerdf,
            center,
            genie_config=genie_config,
        )

        assert expected_results == validate_results
        patch_validate.assert_called_once_with(
            oncotree_link=oncotree_link, nosymbol_check=False
        )
        patch_check.assert_called_once_with(
            validation_statusdf, error_trackerdf, entities
        )
        patch_determine_filetype.assert_called_once()
        patch_staterror_list.assert_called_once_with(
            valid_results.is_valid(), message, entities
        )
        patch_send_email.assert_not_called()


def test_invalid_validatefile(genie_config):
    """
    Tests the behavior of a file that gets validated that becomes
    invalid
    """
    validation_statusdf = pd.DataFrame()
    error_trackerdf = pd.DataFrame(columns=["id"], dtype=str)
    entity = synapseclient.Entity(
        name="data_clinical_supp_SAGE.txt",
        id="syn1234",
        md5="44444",
        path="/path/to/data_clinical_supp_SAGE.txt",
    )
    entity["modifiedOn"] = "2019-03-24T12:00:00.Z"
    # This modifiedOn translates to: 1553428800000
    entity.modifiedBy = "333"
    entity.createdBy = "444"
    entities = [entity]
    valid_results = example_filetype_format.ValidationResults(
        errors="Is invalid", warnings=""
    )
    message = "Is invalid"
    filetype = "clinical"
    status_error_list_results = (
        [{"entity": entity, "status": "INVALID"}],
        [{"entity": entity, "errors": message}],
    )
    expected_results = (
        [
            {
                "entity": entity,
                "status": "INVALID",
                "fileType": filetype,
                "center": center,
            }
        ],
        [{"entity": entity, "errors": message, "fileType": filetype, "center": center}],
        [(["data_clinical_supp_SAGE.txt"], "Is invalid", ["333", "444"])],
    )

    with patch.object(
        GenieValidationHelper, "determine_filetype", return_value=filetype
    ) as patch_determine_filetype, patch.object(
        input_to_database,
        "check_existing_file_status",
        return_value={"status_list": [], "error_list": [], "to_validate": True},
    ) as patch_check, patch.object(
        GenieValidationHelper,
        "validate_single_file",
        return_value=(valid_results, message),
    ) as patch_validate, patch.object(
        input_to_database,
        "_get_status_and_error_list",
        return_value=status_error_list_results,
    ) as patch_staterror_list:

        validate_results = input_to_database.validatefile(
            syn,
            None,
            entities,
            validation_statusdf,
            error_trackerdf,
            center,
            genie_config=genie_config,
        )

        assert expected_results == validate_results
        patch_validate.assert_called_once_with(
            oncotree_link=oncotree_link, nosymbol_check=False
        )
        patch_check.assert_called_once_with(
            validation_statusdf, error_trackerdf, entities
        )
        patch_determine_filetype.assert_called_once()
        patch_staterror_list.assert_called_once_with(
            valid_results.is_valid(), message, entities
        )


def test_already_validated_validatefile():
    """
    Test already validated files
    """
    validation_statusdf = pd.DataFrame()
    error_trackerdf = pd.DataFrame()
    entity = synapseclient.Entity(
        name="data_clinical_supp_SAGE.txt",
        id="syn1234",
        md5="44444",
        path="/path/to/data_clinical_supp_SAGE.txt",
    )
    entity["modifiedOn"] = "2019-03-24T12:00:00.Z"
    # This modifiedOn translates to: 1553428800000
    entity.modifiedBy = "333"
    entity.createdBy = "444"
    entities = [entity]
    threads = 0
    valid = False
    errors = "Invalid file"
    filetype = "markdown"
    status = "INVALID"
    check_file_status_dict = {
        "status_list": [status],
        "error_list": [errors],
        "to_validate": False,
    }

    status_error_list_results = (
        [{"entity": entity, "status": status}],
        [{"entity": entity, "errors": errors}],
    )
    expected_results = (
        [{"entity": entity, "status": status, "fileType": filetype, "center": center}],
        [{"entity": entity, "errors": errors, "fileType": filetype, "center": center}],
        [],
    )
    with patch.object(
        GenieValidationHelper, "determine_filetype", return_value=filetype
    ) as patch_determine_filetype, patch.object(
        input_to_database,
        "check_existing_file_status",
        return_value=check_file_status_dict,
    ) as patch_check, patch.object(
        GenieValidationHelper, "validate_single_file", return_value=(valid, errors)
    ) as patch_validate, patch.object(
        input_to_database,
        "_get_status_and_error_list",
        return_value=status_error_list_results,
    ) as patch_staterror_list, patch.object(
        input_to_database, "_send_validation_error_email"
    ) as patch_send_email:

        validate_results = input_to_database.validatefile(
            syn,
            None,
            entities,
            validation_statusdf,
            error_trackerdf,
            center,
            threads,
            oncotree_link,
        )

        assert expected_results == validate_results
        patch_validate.assert_not_called()
        patch_check.assert_called_once_with(
            validation_statusdf, error_trackerdf, entities
        )
        patch_determine_filetype.assert_called_once()
        patch_staterror_list.assert_not_called()
        patch_send_email.assert_not_called()


def test_valid__get_status_and_error_list():
    """
    Tests the correct status and error lists received
    when file is valid.
    """
    # modified_on = 1561143558000
    modified_on_string = "2019-06-21T18:59:18.456Z"

    entity = synapseclient.Entity(
        id="syn1234",
        md5="44444",
        path="/path/to/foobar.txt",
        name="data_clinical_supp_SAGE.txt",
    )
    entity.properties.modifiedOn = modified_on_string

    entities = [entity]

    valid = True
    message = "valid"
    # filetype = 'clinical'

    (
        input_status_list,
        invalid_errors_list,
    ) = input_to_database._get_status_and_error_list(valid, message, entities)
    assert input_status_list == [{"entity": entity, "status": "VALIDATED"}]
    assert not invalid_errors_list


def test_invalid__get_status_and_error_list():
    """
    Tests the correct status and error lists received
    when file is invalid.
    """
    # modified_on = 1561143558000
    modified_on_string = "2019-06-21T18:59:18.456Z"
    entity = synapseclient.Entity(
        id="syn1234",
        md5="44444",
        path="/path/to/foobar.txt",
        name="data_clinical_supp_SAGE.txt",
    )
    entity.properties.modifiedOn = modified_on_string

    entities = [entity]
    # filetype = "clinical"
    # This valid variable control the validation status
    valid = False
    errors = "invalid file content"

    (
        input_status_list,
        invalid_errors_list,
    ) = input_to_database._get_status_and_error_list(valid, errors, entities)
    assert input_status_list == [{"entity": entity, "status": "INVALID"}]
    assert invalid_errors_list == [{"entity": entity, "errors": errors}]


def test__send_validation_error_email():
    message = "invalid error message here"
    filenames = ["data_clinical_supp_SAGE.txt"]
    file_user = "333"
    message_objs = [dict(filenames=filenames, messages=message)]
    with patch.object(
        syn, "getUserProfile", return_value={"userName": "trial"}
    ) as patch_syn_getuserprofile, patch.object(
        syn, "sendMessage"
    ) as patch_syn_sendmessage, patch(
        "genie.input_to_database.datetime"
    ) as mock_datetime:
        mock_datetime.datetime.today.return_value = datetime(2019, 11, 1, 00, 00, 00, 0)
        mock_datetime.side_effect = lambda *args, **kw: date(*args, **kw)

        input_to_database._send_validation_error_email(syn, file_user, message_objs)
        error_message = (
            "Dear trial,\n\n"
            "You have invalid files! "
            "Here are the reasons why:\n\n"
            "Filenames: data_clinical_supp_SAGE.txt, "
            f"Errors:\n {message}\n\n"
        )
        patch_syn_sendmessage.assert_called_once_with(
            userIds=[file_user],
            messageBody=error_message,
            messageSubject="GENIE Validation Error - 2019-11-01 00:00:00",
        )
        patch_syn_getuserprofile.assert_called_once_with(file_user)


class emptytable_mock:
    """
    Validation status tablequery dataframe mocking and
    error tracking tablequery dataframe mocking
    This is used because assert_called_once_with has a hard
    time with comparing pandas dataframes
    """

    tableId = "syn555"

    def asDataFrame(self):
        return []


class TestValidation:
    def setup(self):
        valid = [
            [
                sample_clinical_entity.id,
                sample_clinical_entity.path,
                sample_clinical_entity.md5,
                "VALIDATED",
                sample_clinical_entity.name,
                1553428800000,
                "clinical",
                center,
                sample_clinical_entity,
            ]
        ]
        self.empty_validation = pd.DataFrame(
            columns=[
                "id",
                "path",
                "md5",
                "status",
                "name",
                "modifiedOn",
                "fileType",
                "center",
                "entity",
            ]
        )
        self.validation_statusdf = pd.DataFrame(
            valid,
            columns=[
                "id",
                "path",
                "md5",
                "status",
                "name",
                "modifiedOn",
                "fileType",
                "center",
                "entity",
            ],
        )
        error = [
            [
                sample_clinical_entity.id,
                "errors",
                sample_clinical_entity.name,
                "clinical",
                center,
                sample_clinical_entity,
            ]
        ]
        self.errors_df = pd.DataFrame(
            error, columns=["id", "errors", "name", "fileType", "center", "entity"]
        )
        self.empty_errors = pd.DataFrame(
            columns=["id", "errors", "name", "fileType", "center", "entity"]
        )

        self.with_dupsdf = pd.DataFrame(
            {
                "id": [
                    sample_clinical_entity.id,
                    "syn2345",
                    "syn5555",
                    "syn1224",
                    "syn34444",
                ],
                "name": [
                    "first.cbs",
                    "second.seg",
                    "data_clinical_supp_1",
                    "data_clinical_supp_2",
                    "data_clinical_supp_3",
                ],
                "center": ["SAGE"] * 5,
                "fileType": ["type"] * 5,
                "entity": ["entity"] * 5,
            }
        )
        self.duplicateddf = self.with_dupsdf.copy()
        self.duplicateddf["errors"] = input_to_database.DUPLICATED_FILE_ERROR

        self.no_dupsdf = pd.DataFrame(
            {
                "id": [
                    sample_clinical_entity.id,
                    "syn2345",
                    "syn5555",
                    "syn1224",
                    "syn34444",
                ],
                "name": [
                    "cbs.txt",
                    "second.seg",
                    "no_clinical.txt",
                    "data_clinical_supp_2",
                    "data_clinical_supp_3",
                ],
                "center": ["SAGE"] * 5,
                "fileType": ["type"] * 5,
                "entity": ["entity"] * 5,
            }
        )
        self.empty_dup = pd.DataFrame(
            columns=["id", "name", "center", "fileType", "entity", "errors"]
        )

    def test_build_validation_status_table(self):
        input_valid_statuses = [
            {
                "entity": sample_clinical_entity,
                "status": "VALIDATED",
                "fileType": "clinical",
                "center": center,
            }
        ]
        validationdf = input_to_database.build_validation_status_table(
            input_valid_statuses
        )
        assert validationdf.equals(self.validation_statusdf)

    def test_build_validation_status_table__empty(self):
        input_valid_statuses = []
        validationdf = input_to_database.build_validation_status_table(
            input_valid_statuses
        )
        assert validationdf.equals(self.empty_validation)

    def test_build_error_tracking_table(self):
        invalid_errors = [
            {
                "entity": sample_clinical_entity,
                "errors": "errors",
                "fileType": "clinical",
                "center": center,
            }
        ]
        errordf = input_to_database.build_error_tracking_table(invalid_errors)
        assert errordf.equals(self.errors_df)

    def test_build_error_tracking_table__empty(self):
        invalid_errors = []
        errordf = input_to_database.build_error_tracking_table(invalid_errors)
        assert errordf.equals(self.empty_errors)

    def test_update_status_and_error_tables(self):
        """Test updating validation status and error table"""
        validation_status_table = emptytable_mock()
        error_tracker_table = emptytable_mock()

        with patch.object(process_functions, "updateDatabase") as mock_update:
            input_to_database.update_status_and_error_tables(
                syn,
                self.validation_statusdf,
                self.errors_df,
                validation_status_table,
                error_tracker_table,
            )
            assert mock_update.call_count == 2

    def test_dups_get_duplicated_files(self):
        """Test get all duplicates: cbs/seg/clinical"""
        dupsdf = input_to_database.get_duplicated_files(self.with_dupsdf)
        assert dupsdf.equals(self.duplicateddf)

    def test_nodups_get_duplicated_files(self):
        """Test no duplicated"""
        dupsdf = input_to_database.get_duplicated_files(self.no_dupsdf)
        assert dupsdf.equals(self.empty_dup)

    def test__update_tables_content(self):
        """Tests duplicates are added to the tables and errors/statues are
        updated
        """
        errorsdf = self.errors_df.copy()
        errorsdf["errors"] = input_to_database.DUPLICATED_FILE_ERROR

        validationdf = self.validation_statusdf
        validationdf["status"] = "INVALID"

        with patch.object(
            input_to_database, "get_duplicated_files", return_value=self.duplicateddf
        ):
            updated_tables = input_to_database._update_tables_content(
                self.validation_statusdf, self.errors_df
            )
        assert updated_tables["duplicated_filesdf"].equals(self.duplicateddf)
        assert updated_tables["error_trackingdf"].equals(errorsdf)
        assert updated_tables["validation_statusdf"].equals(validationdf)

    def test__update_tables_content__remove_old_duplicates(self):
        """Tests that old duplicates are removed"""
        errorsdf = self.errors_df.copy()
        errorsdf["errors"] = input_to_database.DUPLICATED_FILE_ERROR
        validationdf = self.validation_statusdf
        validationdf["status"] = "INVALID"

        with patch.object(
            input_to_database, "get_duplicated_files", return_value=self.empty_dup
        ):
            updated_tables = input_to_database._update_tables_content(
                validationdf, errorsdf
            )
        assert updated_tables["duplicated_filesdf"].empty
        assert updated_tables["error_trackingdf"].empty
        assert updated_tables["validation_statusdf"].empty

    def test__update_tables_content__remove_old_errors(self):
        """Tests that old errors are removed"""
        errorsdf = self.errors_df.copy()
        errorsdf["errors"] = input_to_database.DUPLICATED_FILE_ERROR
        validationdf = self.validation_statusdf
        validationdf["status"] = "VALIDATED"

        with patch.object(
            input_to_database, "get_duplicated_files", return_value=self.empty_dup
        ):
            updated_tables = input_to_database._update_tables_content(
                validationdf, errorsdf
            )
        assert updated_tables["duplicated_filesdf"].empty
        assert updated_tables["error_trackingdf"].empty
        assert updated_tables["validation_statusdf"].empty

    def test_validation(self, genie_config):
        """Test validation steps"""
        modified_on = 1561143558000
        process = "main"
        entity = synapseclient.Entity(
            id="syn1234",
            md5="44444",
            path="/path/to/foobar.txt",
            name="data_clinical_supp_SAGE.txt",
        )
        entities = [entity]
        filetype = "clinical"
        input_status_list = [
            [
                entity.id,
                entity.path,
                entity.md5,
                "VALIDATED",
                entity.name,
                modified_on,
                filetype,
                center,
            ]
        ]
        invalid_errors_list = []
        messages = []
        new_tables = {
            "validation_statusdf": self.validation_statusdf,
            "error_trackingdf": self.errors_df,
            "duplicated_filesdf": self.empty_dup,
        }
        validationstatus_mock = emptytable_mock()
        errortracking_mock = emptytable_mock()
        valiate_cls = Mock()
        with patch.object(
            syn, "tableQuery", side_effect=[validationstatus_mock, errortracking_mock]
        ) as patch_query, patch.object(
            input_to_database,
            "validatefile",
            return_value=(input_status_list, invalid_errors_list, messages),
        ) as patch_validatefile, patch.object(
            input_to_database,
            "build_validation_status_table",
            return_value=self.validation_statusdf,
        ), patch.object(
            input_to_database, "build_error_tracking_table", return_value=self.errors_df
        ), patch.object(
            input_to_database, "_update_tables_content", return_value=new_tables
        ), patch.object(
            input_to_database, "update_status_and_error_tables"
        ):

            valid_filedf = input_to_database.validation(
                syn,
                "syn123",
                center,
                process,
                entities,
                format_registry={"test": valiate_cls},
                genie_config=genie_config,
            )
            assert patch_query.call_count == 2
            patch_validatefile.assert_called_once_with(
                syn=syn,
                project_id="syn123",
                entities=entity,
                validation_status_table=validationstatus_mock,
                error_tracker_table=errortracking_mock,
                center="SAGE",
                format_registry={"test": valiate_cls},
                genie_config=genie_config,
            )

            assert valid_filedf.equals(
                self.validation_statusdf[["id", "path", "fileType", "name"]]
            )


@pytest.mark.parametrize(
    "process, genieclass, filetype",
    [
        ("main", Mock(), "clinical"),
        ("main", Mock(), "maf"),
    ],
)
def test_main_processfile(genie_config, process, genieclass, filetype):
    validfiles = {
        "id": ["syn1"],
        "path": ["/path/to/data_clinical_supp.txt"],
        "fileType": [filetype],
        "name": ["data_clinical_supp_SAGE.txt"],
    }
    validfilesdf = pd.DataFrame(validfiles)
    center = "SAGE"
    path_to_genie = "./"
    format_registry = {filetype: genieclass}

    input_to_database.processfiles(
        syn,
        validfilesdf,
        center,
        path_to_genie,
        processing=process,
        format_registry=format_registry,
        genie_config=genie_config,
    )
    genieclass.assert_called_once()


def test_mainnone_processfile(genie_config):
    """If file type is None, the processing function is not called"""
    validfiles = {
        "id": ["syn1"],
        "path": ["/path/to/data_clinical_supp_SAGE.txt"],
        "fileType": [None],
        "name": ["data_clinical_supp_SAGE.txt"],
    }
    validfilesdf = pd.DataFrame(validfiles)
    center = "SAGE"
    path_to_genie = "./"
    process_cls = Mock()

    with patch.object(Clinical, "process") as patch_clin:
        input_to_database.processfiles(
            syn,
            validfilesdf,
            center,
            path_to_genie,
            processing="main",
            format_registry={"main": process_cls},
            genie_config=genie_config,
        )
        patch_clin.assert_not_called()


def test_mutation_processfile(genie_config):
    """
    Make sure mutation is called correctly
    """
    validfiles = {
        "id": ["syn1"],
        "path": ["/path/to/data_clinical_supp_SAGE.txt"],
        "fileType": [None],
    }
    validfilesdf = pd.DataFrame(validfiles)
    center = "SAGE"
    path_to_genie = "./"
    process_cls = Mock()

    with patch.object(process_mutation, "process_mutation_workflow") as patch_process:
        input_to_database.processfiles(
            syn,
            validfilesdf,
            center,
            path_to_genie,
            processing="mutation",
            format_registry={"vcf": process_cls},
            genie_config=genie_config,
        )
        patch_process.assert_called_once_with(
            syn=syn,
            center=center,
            validfiles=validfilesdf,
            genie_config=genie_config,
            workdir=path_to_genie,
        )



"""Updates dashboard tables"""
import argparse
import datetime
import logging
import os

import pandas as pd
import synapseclient
from synapseclient.core.utils import to_unix_epoch_time

from genie import process_functions

logger = logging.getLogger(__name__)


def get_center_data_completion(center, df):
    """
    Gets center data completion.  Calulates the percentile of
    how complete a clinical data element is:
    Number of not blank/Unknown/NA divded by
    total number of patients or samples

    Args:
        center: GENIE center
        df: sample or patient dataframe

    Returns:
        Dataframe: Center data
    """
    centerdf = df[df["CENTER"] == center]
    total = len(centerdf)
    center_data = pd.DataFrame()
    skip_cols = [
        "CENTER",
        "PATIENT_ID",
        "SAMPLE_ID",
        "SAMPLE_TYPE_DETAILED",
        "SECONDARY_RACE",
        "TERTIARY_RACE",
    ]
    for col in centerdf:
        if col not in skip_cols:
            not_missing = [
                not pd.isnull(value) and value != "Not Collected"
                for value in centerdf[col]
            ]
            completeness = float(sum(not_missing)) / int(total)
            returned = pd.DataFrame([[col, center, total, completeness]])
            center_data = pd.concat([center_data, returned])
    return center_data


def update_samples_in_release_table(
    syn, file_mapping, release, samples_in_release_synid
):
    """
    Updates the sample in release table
    This tracks the samples of each release.  1 means it exists, and 0
    means it doesn't

    Args:
        syn: synapse object
        file_mapping: file mapping generated from file mapping function
        release:  GENIE release number (ie. 5.3-consortium)
        samples_in_release_synid: Synapse Id of 'samples in release' Table
    """
    clinical_ent = syn.get(file_mapping["clinical"], followLink=True)
    clinicaldf = pd.read_csv(clinical_ent.path, sep="\t", comment="#")
    cols = [i["name"] for i in list(syn.getTableColumns(samples_in_release_synid))]

    if release not in cols:
        schema = syn.get(samples_in_release_synid)
        syn_col = synapseclient.Column(
            name=release, columnType="INTEGER", defaultValue=0
        )
        new_column = syn.store(syn_col)
        schema.addColumn(new_column)
        # no need to return schema variable because it isn't used
        syn.store(schema)
    # Columns of samples in release
    samples_per_release = syn.tableQuery(
        'SELECT SAMPLE_ID, "{}" FROM {}'.format(release, samples_in_release_synid)
    )

    samples_per_releasedf = samples_per_release.asDataFrame()
    new_samples = clinicaldf[["SAMPLE_ID"]][
        ~clinicaldf.SAMPLE_ID.isin(samples_per_releasedf.SAMPLE_ID)
    ]

    new_samples[release] = 1
    old_samples = clinicaldf[["SAMPLE_ID"]][
        clinicaldf.SAMPLE_ID.isin(samples_per_releasedf.SAMPLE_ID)
    ]

    old_samples[release] = 1
    samples_in_releasedf = new_samples.append(old_samples)
    process_functions.updateDatabase(
        syn,
        samples_per_releasedf,
        samples_in_releasedf,
        samples_in_release_synid,
        ["SAMPLE_ID"],
    )


def update_cumulative_sample_table(
    syn, file_mapping, release, cumulative_sample_count_synid
):
    """
    Consortium release sample count table update function
    This gets the cumulative sample count of each file type in each release

    Args:
        syn: synapse object
        file_mapping: file mapping generated from file mapping function
        release:  GENIE release number (ie. 5.3-consortium)
        cumulative_sample_count_synid: Synapse Id of
            'Cumulative sample count' Table
    """

    sample_count_per_round = syn.tableQuery(
        "SELECT * FROM {} where Release = '{}'".format(
            cumulative_sample_count_synid, release
        )
    )
    sample_count_per_rounddf = sample_count_per_round.asDataFrame()

    clinical_ent = syn.get(file_mapping["clinical"], followLink=True)
    clinicaldf = pd.read_csv(clinical_ent.path, sep="\t", comment="#")
    clinicaldf.columns = [i.upper() for i in clinicaldf.columns]
    if clinicaldf.get("CENTER") is None:
        clinicaldf["CENTER"] = [sample.split("-")[1] for sample in clinicaldf.SAMPLE_ID]
    clinical_counts = clinicaldf["CENTER"].value_counts()
    clinical_counts["Total"] = sum(clinical_counts)
    clinical_counts.name = "Clinical"

    fusion_ent = syn.get(file_mapping["fusion"], followLink=True)
    fusiondf = pd.read_csv(fusion_ent.path, sep="\t", comment="#")
    fusiondf.columns = [i.upper() for i in fusiondf.columns]
    fusion_counts = fusiondf["CENTER"][
        ~fusiondf["TUMOR_SAMPLE_BARCODE"].duplicated()
    ].value_counts()
    fusion_counts["Total"] = sum(fusion_counts)

    cna_ent = syn.get(file_mapping["cna"], followLink=True)
    cnadf = pd.read_csv(cna_ent.path, sep="\t", comment="#")
    cna_counts = pd.Series([i.split("-")[1] for i in cnadf.columns[1:]]).value_counts()
    cna_counts["Total"] = sum(cna_counts)

    seg_ent = syn.get(file_mapping["seg"], followLink=True)
    segdf = pd.read_csv(seg_ent.path, sep="\t", comment="#")
    segdf.columns = [i.upper() for i in segdf.columns]

    segdf["CENTER"] = [i.split("-")[1] for i in segdf["ID"]]
    seg_counts = segdf["CENTER"][~segdf["ID"].duplicated()].value_counts()
    seg_counts["Total"] = sum(seg_counts)

    total_counts = pd.DataFrame(clinical_counts)
    total_counts["Fusions"] = fusion_counts
    total_counts["CNV"] = cna_counts
    total_counts["Mutation"] = clinical_counts
    total_counts["SEG"] = seg_counts
    total_counts = total_counts.fillna(0)
    total_counts = total_counts.applymap(int)
    total_counts["Center"] = total_counts.index
    total_counts["Release"] = release
    process_functions.updateDatabase(
        syn,
        sample_count_per_rounddf,
        total_counts,
        cumulative_sample_count_synid,
        ["Center", "Release"],
        to_delete=True,
    )


def get_file_mapping(syn, release_folder_synid):
    """
    Get file mapping between important files needed for dashboard and
    their synapse ids

    Args:
        syn:  synapse object
        release_folder_synid: synapse id of release

    """
    files = syn.getChildren(release_folder_synid)
    file_mapping = dict()
    for metadata in files:
        filename = metadata["name"]
        synid = metadata["id"]
        if not filename.startswith("meta"):
            if filename.startswith("data_clinical_sample"):
                file_mapping["clinical"] = synid
            elif filename.endswith("fusions.txt"):
                file_mapping["fusion"] = synid
            elif filename.endswith("CNA.txt"):
                file_mapping["cna"] = synid
            elif filename.endswith(".seg"):
                file_mapping["seg"] = synid
    return file_mapping


def update_release_numbers(syn, database_mappingdf, release=None):
    """
    Updates all release dashboard numbers or specific release number

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
        release: GENIE release (ie. 5.3-consortium).  Defaults to None
    """
    # Update release table with current release or all releases
    samples_in_release_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "samplesInRelease"
    ].values[0]
    cumulative_sample_count_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "cumulativeSampleCount"
    ].values[0]

    release_folder_fileview_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "releaseFolder"
    ].values[0]
    release_folder = syn.tableQuery(
        "select id,name from %s" % release_folder_fileview_synid
        + " where name not like 'Release%' and name <> 'case_lists' and "
        + "name not like '%.0.%'"
    )
    release_folderdf = release_folder.asDataFrame()

    for rel_synid, rel_name in zip(release_folderdf.id, release_folderdf.name):
        file_mapping = get_file_mapping(syn, rel_synid)
        # If release is specified, only process on that,
        # otherwise process for all
        if release is None or release == rel_name:
            update_samples_in_release_table(
                syn, file_mapping, rel_name, samples_in_release_synid
            )
            update_cumulative_sample_table(
                syn, file_mapping, rel_name, cumulative_sample_count_synid
            )
        else:
            pass


def update_database_numbers(syn, database_mappingdf):
    """
    Updates database cumulative numbers (Only called when not staging)

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    cumulative_sample_count_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "cumulativeSampleCount"
    ].values[0]
    # Database
    database_count = syn.tableQuery(
        "SELECT * FROM %s where Release = 'Database'" % cumulative_sample_count_synid
    )
    database_countdf = database_count.asDataFrame()
    clinical = syn.tableQuery("select CENTER from syn7517674")
    clinicaldf = clinical.asDataFrame()
    clinincal_counts = clinicaldf["CENTER"].value_counts()
    clinincal_counts["Total"] = sum(clinincal_counts)
    clinincal_counts.name = "Clinical"

    fusion = syn.tableQuery("select * from syn7893268")
    fusiondf = fusion.asDataFrame()
    fusion_counts = fusiondf["CENTER"][
        ~fusiondf["TUMOR_SAMPLE_BARCODE"].duplicated()
    ].value_counts()
    fusion_counts["Total"] = sum(fusion_counts)

    center_flat_files = syn.getChildren("syn12278118")
    cna_file_paths = [
        syn.get(file["id"]).path
        for file in center_flat_files
        if file["name"].startswith("data_CNA")
    ]
    cna_numbers = {}
    for cna_file in cna_file_paths:
        center = os.path.basename(cna_file).replace(".txt", "").split("_")[2]
        with open(cna_file, "r") as cna:
            header = cna.readline()
            samples = header.split("\t")
            # Minus one because of Hugo_Symbol
            cna_numbers[center] = len(samples) - 1
    cna_counts = pd.Series(cna_numbers)
    cna_counts["Total"] = sum(cna_counts)

    seg = syn.tableQuery("select * from syn7893341")
    segdf = seg.asDataFrame()
    seg_counts = segdf["CENTER"][~segdf["ID"].duplicated()].value_counts()
    seg_counts["Total"] = sum(seg_counts)

    db_counts = pd.DataFrame(clinincal_counts)
    db_counts["Fusions"] = fusion_counts
    db_counts["CNV"] = cna_counts
    db_counts["Mutation"] = clinincal_counts
    db_counts["SEG"] = seg_counts
    db_counts = db_counts.fillna(0)
    db_counts = db_counts.applymap(int)
    db_counts["Center"] = db_counts.index
    db_counts["Release"] = "Database"
    process_functions.updateDatabase(
        syn,
        database_countdf,
        db_counts,
        cumulative_sample_count_synid,
        ["Center", "Release"],
    )
    today = datetime.date.today()
    if today.month in [1, 4, 8, 12]:
        db_count_tracker = db_counts[["Clinical", "Center", "Release"]]
        db_count_tracker.rename(
            columns={"Clinical": "sample_count", "Center": "center", "Release": "date"},
            inplace=True,
        )
        db_count_tracker["date"] = today.strftime("%b-%Y")
        # Hard coded syn id
        syn.store(synapseclient.Table("syn18404852", db_count_tracker))


def update_oncotree_code_tables(syn, database_mappingdf):
    """
    Updates database statistics of oncotree codes
    and primary onocotree codes

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    oncotree_distribution_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "oncotree"
    ].values[0]

    clinical = syn.tableQuery("select * from syn7517674")
    clinicaldf = clinical.asDataFrame()

    # DISTRIBUTION OF ONCOTREE CODE TABLE UPDATE
    oncotree_code_distributiondf = pd.DataFrame(
        columns=set(clinicaldf["CENTER"]), index=set(clinicaldf["ONCOTREE_CODE"])
    )
    for center in oncotree_code_distributiondf.columns:
        onc_counts = clinicaldf["ONCOTREE_CODE"][
            clinicaldf["CENTER"] == center
        ].value_counts()
        oncotree_code_distributiondf[center] = onc_counts
    oncotree_code_distributiondf = oncotree_code_distributiondf.fillna(0)
    oncotree_code_distributiondf = oncotree_code_distributiondf.applymap(int)
    oncotree_code_distributiondf["Total"] = oncotree_code_distributiondf.apply(
        sum, axis=1
    )
    oncotree_code_distributiondf["Oncotree_Code"] = oncotree_code_distributiondf.index

    oncotree_distribution_db = syn.tableQuery(
        "SELECT %s FROM %s"
        % (
            "Oncotree_Code," + ",".join(clinicaldf["CENTER"].unique()) + ",Total",
            oncotree_distribution_synid,
        )
    )

    oncotree_distribution_dbdf = oncotree_distribution_db.asDataFrame()
    process_functions.updateDatabase(
        syn,
        oncotree_distribution_dbdf,
        oncotree_code_distributiondf,
        oncotree_distribution_synid,
        ["Oncotree_Code"],
        to_delete=True,
    )

    # DISTRIBUTION OF PRIMARY CODE TABLE UPDATE
    oncotree_link_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "oncotreeLink"
    ].values[0]
    primary_code_synId = database_mappingdf["Id"][
        database_mappingdf["Database"] == "primaryCode"
    ].values[0]

    # Can also use most up to date oncotree code,
    # because these tables are updated from the database
    oncotree_link_ent = syn.get(oncotree_link_synid)
    oncotree_link = oncotree_link_ent.externalURL
    oncotree_mapping = process_functions.get_oncotree_code_mappings(oncotree_link)

    clinicaldf["PRIMARY_CODES"] = [
        oncotree_mapping[i.upper()]["ONCOTREE_PRIMARY_NODE"]
        if i.upper() in oncotree_mapping.keys()
        else "DEPRECATED_CODE"
        for i in clinicaldf.ONCOTREE_CODE
    ]

    # ### DISTRIBUTION OF PRIMARY ONCOTREE CODE TABLE UPDATE
    primary_code_distributiondf = pd.DataFrame(
        columns=set(clinicaldf["CENTER"]), index=set(clinicaldf["PRIMARY_CODES"])
    )

    for center in primary_code_distributiondf.columns:
        onc_counts = clinicaldf["PRIMARY_CODES"][
            clinicaldf["CENTER"] == center
        ].value_counts()
        primary_code_distributiondf[center] = onc_counts
    primary_code_distributiondf = primary_code_distributiondf.fillna(0)
    primary_code_distributiondf = primary_code_distributiondf.applymap(int)
    primary_code_distributiondf["Total"] = primary_code_distributiondf.apply(
        sum, axis=1
    )
    primary_code_distributiondf["Oncotree_Code"] = primary_code_distributiondf.index

    primary_code_dist_db = syn.tableQuery(
        "SELECT %s FROM %s"
        % (
            "Oncotree_Code," + ",".join(clinicaldf["CENTER"].unique()) + ",Total",
            primary_code_synId,
        )
    )

    primary_code_dist_dbdf = primary_code_dist_db.asDataFrame()
    process_functions.updateDatabase(
        syn,
        primary_code_dist_dbdf,
        primary_code_distributiondf,
        primary_code_synId,
        ["Oncotree_Code"],
        to_delete=True,
    )


def update_sample_difference_table(syn, database_mappingdf):
    """
    Updates sample difference table between
    consortium releases

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    cumulative_sample_count_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "cumulativeSampleCount"
    ].values[0]

    sample_diff_count_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "sampleDiffCount"
    ].values[0]

    # UPDATE DIFF TABLE
    sample_count_per_round = syn.tableQuery(
        "SELECT * FROM %s where Center <> 'Total' and Release <> 'Database'"
        % cumulative_sample_count_synid
    )

    sample_count_per_rounddf = sample_count_per_round.asDataFrame()
    releases = list(sample_count_per_rounddf["Release"].unique())
    # sort the releases and remove public releases
    releases.sort()
    consortium_releases = [
        release
        for release in releases
        if "public" not in release and ".0." not in release
    ]

    diff_between_releasesdf = sample_count_per_rounddf[
        sample_count_per_rounddf["Release"] == consortium_releases[0]
    ]

    for index, release_name in enumerate(consortium_releases[1:]):
        prior_release = sample_count_per_rounddf[
            sample_count_per_rounddf["Release"] == consortium_releases[index]
        ]

        current_release = sample_count_per_rounddf[
            sample_count_per_rounddf["Release"] == release_name
        ]

        prior_release.index = prior_release["Center"]
        current_release.index = current_release["Center"]

        del prior_release["Center"]
        del prior_release["Release"]
        del current_release["Center"]
        del current_release["Release"]
        # Append new rows of centers that are new and
        # just added to the releases
        new_centers = current_release.index[
            ~current_release.index.isin(prior_release.index)
        ]

        if not new_centers.empty:
            prior_release = pd.concat([prior_release, pd.DataFrame(index=new_centers)])
            prior_release = prior_release.fillna(0)
        difference = current_release - prior_release
        difference["Center"] = difference.index
        difference["Release"] = release_name
        diff_between_releasesdf = pd.concat([diff_between_releasesdf, difference])

    difftable_db = syn.tableQuery("SELECT * FROM %s" % sample_diff_count_synid)
    difftable_dbdf = difftable_db.asDataFrame()
    difftable_dbdf = difftable_dbdf.fillna(0)
    new_values = (
        diff_between_releasesdf[["Clinical", "Mutation", "CNV", "SEG", "Fusions"]]
        .fillna(0)
        .applymap(int)
    )

    diff_between_releasesdf[
        ["Clinical", "Mutation", "CNV", "SEG", "Fusions"]
    ] = new_values

    process_functions.updateDatabase(
        syn,
        difftable_dbdf,
        diff_between_releasesdf,
        sample_diff_count_synid,
        ["Center", "Release"],
        to_delete=True,
    )


def update_data_completeness_table(syn, database_mappingdf):
    """
    Updates the data completeness of the database

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    data_completion_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "dataCompletion"
    ].values[0]

    sample = syn.tableQuery("select * from syn7517674")
    sampledf = sample.asDataFrame()
    patient = syn.tableQuery("select * from syn7517669")
    patientdf = patient.asDataFrame()

    data_completenessdf = pd.DataFrame()
    center_infos = sampledf.CENTER.drop_duplicates().apply(
        lambda center: get_center_data_completion(center, sampledf)
    )
    for center_info in center_infos:
        data_completenessdf = pd.concat([data_completenessdf, center_info])

    center_infos = patientdf.CENTER.drop_duplicates().apply(
        lambda center: get_center_data_completion(center, patientdf)
    )
    for center_info in center_infos:
        data_completenessdf = pd.concat([data_completenessdf, center_info])

    data_completeness_db = syn.tableQuery("select * from %s" % data_completion_synid)
    data_completeness_dbdf = data_completeness_db.asDataFrame()
    data_completenessdf.columns = data_completeness_dbdf.columns
    process_functions.updateDatabase(
        syn,
        data_completeness_dbdf,
        data_completenessdf,
        data_completion_synid,
        ["FIELD", "CENTER"],
        to_delete=True,
    )


def update_wiki(syn, database_mappingdf):
    """
    Updates the GENIE project dashboard wiki timestamp

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database

    """
    # Updates to query and date dashboard was updated
    cumulative_sample_count_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "cumulativeSampleCount"
    ].values[0]

    primary_code_synId = database_mappingdf["Id"][
        database_mappingdf["Database"] == "primaryCode"
    ].values[0]

    centers = syn.tableQuery("select distinct(CENTER) as CENTER from syn7517674")

    centersdf = centers.asDataFrame()
    now = datetime.datetime.now()
    markdown = [
        "_Updated {month}/{day}/{year}_\n\n".format(
            month=now.month, day=now.day, year=now.year
        ),
        "##Count of Clinical Samples\n",
        "${synapsetable?query=SELECT Center%2C Clinical%2C Release FROM "
        + cumulative_sample_count_synid
        + "}\n\n",
        "\n\n##Primary Oncotree Codes\n\n",
        "${synapsetable?query=SELECT Oncotree%5FCode%2C "
        + "%2C ".join(centersdf["CENTER"].unique())
        + "%2C Total FROM "
        + primary_code_synId
        + " ORDER BY Total DESC&limit=15}\n\n",
    ]

    wikipage = syn.getWiki("syn3380222", 235803)
    wikipage.markdown = "".join(markdown)
    syn.store(wikipage)


def string_to_unix_epoch_time_milliseconds(string_time):
    """
    Takes dates in this format: 2018-10-25T20:16:07.959Z
    and turns it into unix epoch time in milliseconds

    Args:
        string_time: string in this format: 2018-10-25T20:16:07.959Z

    Returns:
        unix epoch time in milliseconds
    """
    datetime_obj = datetime.datetime.strptime(
        string_time.split(".")[0], "%Y-%m-%dT%H:%M:%S"
    )
    return to_unix_epoch_time(datetime_obj)


def update_data_release_file_table(syn, database_mappingdf):
    """
    Updates data release file table

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    release_folder_fileview_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "releaseFolder"
    ].values[0]
    # Add this query so that the fileview is indexed to include the
    # new release folder
    syn.tableQuery(f"select * from {release_folder_fileview_synid} limit 1")
    release_folder = syn.tableQuery(
        "select id,name from %s" % release_folder_fileview_synid
        + " where name not like 'Release%' and name <> 'case_lists' "
        + "and name not like '0.%'"
    )
    release_folderdf = release_folder.asDataFrame()

    data_release_table_synid = "syn16804261"
    data_release_table = syn.tableQuery("select * from %s" % data_release_table_synid)
    data_release_tabledf = data_release_table.asDataFrame()

    not_in_release_tabledf = release_folderdf[
        ~release_folderdf.name.isin(data_release_tabledf.release)
    ]

    for synid, name in zip(not_in_release_tabledf.id, not_in_release_tabledf.name):
        release_files = syn.getChildren(synid)

        append_rows = [
            [
                release_file["name"],
                release_file["id"],
                name,
                string_to_unix_epoch_time_milliseconds(release_file["modifiedOn"]),
                synid,
            ]
            for release_file in release_files
            if release_file["name"] != "case_lists"
        ]

        syn.store(synapseclient.Table(data_release_table_synid, append_rows))


def check_column_decreases(currentdf, olderdf):
    """
    Checks entity decreases

    Args:
        current_ent: Current entity dataframe
        old_ent: Older entity dataframe

    Returns:
        Differences in values
    """
    diff_map = dict()
    for col in currentdf:
        new_counts = currentdf[col].value_counts()
        if olderdf.get(col) is not None:
            old_counts = olderdf[col].value_counts()
            # Make sure any values that exist in the new get added
            # to the old to show the decrease
            new_keys = pd.Series(
                index=new_counts.keys()[~new_counts.keys().isin(old_counts.keys())]
            )
            old_counts = old_counts.add(new_keys, fill_value=0)
            old_counts.fillna(0, inplace=True)
            # Make sure any values that don't exist in the old get added
            # to show the decrease
            new_keys = pd.Series(
                index=old_counts.keys()[~old_counts.keys().isin(new_counts.keys())]
            )
            new_counts = new_counts.add(new_keys, fill_value=0)
            new_counts.fillna(0, inplace=True)
            if any(new_counts - old_counts < 0):
                logger.info("\tDECREASE IN COLUMN: %s" % col)
                # diff = new_counts[new_counts - old_counts < 0]
                diffs = new_counts - old_counts
                diffstext = diffs[diffs < 0].to_csv().replace("\n", "; ")
                logger.info("\t" + diffstext)
                diff_map[col] = True
            else:
                diff_map[col] = False
    return diff_map


def print_clinical_values_difference_table(syn, database_mappingdf):
    """
    Checks for a decrease in values in the clinical file
    from last consortium release to most recent consortium release

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
    """
    release_folder_fileview_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "releaseFolder"
    ].values[0]

    clinical_key_decrease_synid = database_mappingdf["Id"][
        database_mappingdf["Database"] == "clinicalKeyDecrease"
    ].values[0]

    release_folder = syn.tableQuery(
        f"select id,name from {release_folder_fileview_synid} "
        "where name not like 'Release%' and name <> 'case_lists' "
        "and name not like '%.0.%' and name not like '%-public' "
        "and name <> 'potential_artifacts'"
    )

    release_folderdf = release_folder.asDataFrame()
    # Set release number as a numerical value since string "10" < "9"
    # Also can't set by created on date, because sometimes
    # there are patch releases
    release_folderdf["num_release"] = [
        float(name.replace(".0", "").replace("-consortium", ""))
        for name in release_folderdf["name"]
    ]
    release_folderdf.sort_values("num_release", ascending=False, inplace=True)
    current_release = release_folderdf["id"][0]
    older_release = release_folderdf["id"][1]

    current_release_files = syn.getChildren(current_release)
    current_clinical_synids = {
        file["name"]: file["id"]
        for file in current_release_files
        if file["name"] in ["data_clinical_sample.txt", "data_clinical_patient.txt"]
    }

    older_release_files = syn.getChildren(older_release)

    older_clinical_synids = {
        file["name"]: file["id"]
        for file in older_release_files
        if file["name"] in ["data_clinical_sample.txt", "data_clinical_patient.txt"]
    }

    current_sample_ent = syn.get(
        current_clinical_synids["data_clinical_sample.txt"], followLink=True
    )

    older_sample_ent = syn.get(
        older_clinical_synids["data_clinical_sample.txt"], followLink=True
    )
    current_sampledf = pd.read_csv(current_sample_ent.path, sep="\t", comment="#")

    current_sampledf["CENTER"] = [
        patient.split("-")[1] for patient in current_sampledf["PATIENT_ID"]
    ]

    older_sampledf = pd.read_csv(older_sample_ent.path, sep="\t", comment="#")
    older_sampledf["CENTER"] = [
        patient.split("-")[1] for patient in older_sampledf["PATIENT_ID"]
    ]
    # Rather than take the CENTER, must take the SAMPLE_ID to compare
    current_sampledf = current_sampledf[
        current_sampledf["SAMPLE_ID"].isin(older_sampledf["SAMPLE_ID"].unique())
    ]

    logger.info("SAMPLE CLINICAL VALUE DECREASES")
    center_decrease_mapping = dict()
    for center in older_sampledf["CENTER"].unique():
        current_center_sampledf = current_sampledf[current_sampledf["CENTER"] == center]

        older_center_sampledf = older_sampledf[older_sampledf["CENTER"] == center]

        logger.info(center)

        decrease_map = check_column_decreases(
            current_center_sampledf, older_center_sampledf
        )
        center_decrease_mapping[center] = decrease_map

    current_patient_ent = syn.get(
        current_clinical_synids["data_clinical_patient.txt"], followLink=True
    )

    older_patient_ent = syn.get(
        older_clinical_synids["data_clinical_patient.txt"], followLink=True
    )

    current_patientdf = pd.read_csv(current_patient_ent.path, sep="\t", comment="#")

    older_patientdf = pd.read_csv(older_patient_ent.path, sep="\t", comment="#")
    # Rather than take the CENTER, must take the PATIENT_ID to compare
    current_patientdf = current_patientdf[
        current_patientdf["PATIENT_ID"].isin(older_patientdf["PATIENT_ID"].unique())
    ]

    logger.info("PATIENT CLINICAL VALUE DECREASES")
    for center in older_patientdf["CENTER"].unique():
        current_center_patientdf = current_patientdf[
            current_patientdf["CENTER"] == center
        ]

        older_center_patientdf = older_patientdf[older_patientdf["CENTER"] == center]

        logger.info(center)
        patient_decrease_map = check_column_decreases(
            current_center_patientdf, older_center_patientdf
        )

        center_decrease_mapping[center].update(patient_decrease_map)

    center_decrease_mapping = pd.DataFrame(center_decrease_mapping)
    center_decrease_mapping = center_decrease_mapping.transpose()
    center_decrease_mapping["CENTER"] = center_decrease_mapping.index

    clinical_key_decrease = syn.tableQuery(
        "select * from {0}".format(clinical_key_decrease_synid)
    )
    clinical_key_decreasedbdf = clinical_key_decrease.asDataFrame()
    process_functions.updateDatabase(
        syn,
        clinical_key_decreasedbdf,
        center_decrease_mapping,
        clinical_key_decrease_synid,
        ["CENTER"],
        to_delete=True,
    )


def run_dashboard(syn, database_mappingdf, release, staging=False, public=False):
    """
    Runs the dashboard scripts

    Args:
        syn: synapse object
        database_mappingdf: mapping between synapse ids and database
        release: GENIE release (ie. 5.3-consortium)

    """
    update_release_numbers(syn, database_mappingdf, release=release)

    if not staging:
        update_data_release_file_table(syn, database_mappingdf)
        if not public:
            print_clinical_values_difference_table(syn, database_mappingdf)
            update_sample_difference_table(syn, database_mappingdf)
            update_data_completeness_table(syn, database_mappingdf)
            update_database_numbers(syn, database_mappingdf)
            update_oncotree_code_tables(syn, database_mappingdf)
            update_wiki(syn, database_mappingdf)


def main():
    parser = argparse.ArgumentParser(description="Update dashboard tables")

    parser.add_argument(
        "--release", help="GENIE release number (ie. 5.3-consortium)", default=None
    )

    parser.add_argument("--pem_file", type=str, help="Path to PEM file (genie.pem)")

    parser.add_argument(
        "--staging", action="store_true", help="Using staging directory files"
    )

    parser.add_argument("--debug", action="store_true", help="Synapse debugging flag")

    parser.add_argument(
        "--public", action="store_true", help="Set true if releasing public release"
    )

    args = parser.parse_args()
    syn = process_functions.synLogin(args)
    if args.staging:
        # Database to Synapse Id mapping Table
        database_mapping_synid = "syn12094210"
    else:
        database_mapping_synid = "syn10967259"

    database_mapping = syn.tableQuery("select * from %s" % database_mapping_synid)
    database_mappingdf = database_mapping.asDataFrame()

    run_dashboard(
        syn, database_mappingdf, args.release, staging=args.staging, public=args.public
    )


if __name__ == "__main__":
    main()



"""Test write invalid reasons module"""
from unittest import mock
from unittest.mock import create_autospec, patch

from genie import write_invalid_reasons
import pandas as pd
import synapseclient


SYN = create_autospec(synapseclient.Synapse)
CENTER_ERRORSDF = pd.DataFrame(
    {
        "id": ["syn1234", "syn2345"],
        "errors": ["my errors|n", "errors here|f"],
        "center": ["SAGE", "TEST"],
    }
)
ENT1 = synapseclient.File(id="syn123", name="test", parentId="syn3333")


class QueryResponse:
    def asDataFrame():
        return CENTER_ERRORSDF


def test__combine_center_file_errors():
    """Test combining each center's file errors"""
    expected_error = (
        f"\t{ENT1.name} ({ENT1.id}):\n\nmy errors\nn\n\n"
        f"\t{ENT1.name} ({ENT1.id}):\n\nerrors here\nf\n\n"
    )
    calls = [
        mock.call("syn1234", downloadFile=False),
        mock.call("syn2345", downloadFile=False),
    ]
    with patch.object(SYN, "get", return_value=ENT1) as patch_synget:
        center_errors = write_invalid_reasons._combine_center_file_errors(
            SYN, CENTER_ERRORSDF
        )
        assert center_errors == expected_error
        patch_synget.assert_has_calls(calls)


def test_get_center_invalid_errors():
    """Test getting all center invalid errors"""
    with patch.object(
        SYN, "tableQuery", return_value=QueryResponse
    ) as patch_query, patch.object(
        write_invalid_reasons, "_combine_center_file_errors", return_value="errors"
    ) as patch_combine:
        center_invalid = write_invalid_reasons.get_center_invalid_errors(SYN, "syn3333")
        assert center_invalid == {"SAGE": "errors", "TEST": "errors"}
        patch_query.assert_called_once_with("SELECT * FROM syn3333")
        assert patch_combine.call_count == 2



#! /usr/bin/env python3
"""Script to crawl Synapse folder for a center, validate, and update database tables.

"""
import argparse
from datetime import date
import logging

from genie import (
    config,
    input_to_database,
    process_functions,
    write_invalid_reasons,
)

logger = logging.getLogger(__name__)


def main(
    process: str,
    project_id: str,
    center=None,
    pemfile=None,
    delete_old=False,
    only_validate=False,
    oncotree_link=None,
    genie_annotation_pkg=None,
    create_new_maf_database=False,
    debug=False,
    format_registry=None,
):
    """Invoke the GENIE ETL pipeline from data input files to synapse tables

    Args:
        process (str): main or mutation processing
        project_id (str): Synapse project id that houses GENIE project
        center (str, optional): GENIE center. Defaults to None.
        pemfile (str, optional): Path to private key. Defaults to None.
        delete_old (bool, optional): True to delete all old input/processed files.
                                     Defaults to False.
        only_validate (bool, optional): True if only validate files. Defaults to False.
        oncotree_link (str, optional): Link to oncotree version. Defaults to None.
        genie_annotation_pkg (str, optional): vcf/maf conversion tools.
                                              Defaults to None.
        create_new_maf_database (bool, optional): To create new maf table.
                                                  Defaults to False.
        debug (bool, optional): Debug mode. Defaults to False.
        format_registry (str, optional): File format registry python package.
                                         Defaults to None.

    Raises:
        ValueError: If invalid center name is specified
        Exception: If processing is already happening.
    """

    syn = process_functions.synLogin(pemfile, debug=debug)

    # Get project GENIE configurations
    genie_config = process_functions.get_genie_config(syn=syn, project_id=project_id)

    # Filter for specific center
    if center is not None:
        if center not in genie_config["center_config"].keys():
            raise ValueError(
                "Must specify one of these centers: {}".format(
                    ", ".join(genie_config["center_config"].keys())
                )
            )
        centers = [center]
    else:
        # TODO: add in logic to exclude sites from processing
        centers = list(genie_config["center_config"].keys())

    # HACK: Modify oncotree link config
    if oncotree_link is None:
        onco_link_ent = syn.get(genie_config["oncotreeLink"])
        oncotree_link = onco_link_ent.externalURL
    genie_config["oncotreeLink"] = oncotree_link
    # Check if you can connect to oncotree link,
    # if not then don't run validation / processing
    process_functions.checkUrl(genie_config["oncotreeLink"])

    # HACK: Add genie annotation package to config
    if process == "mutation" and genie_annotation_pkg is None:
        raise ValueError("Must define genie annotation pkg if mutation processing")
    genie_config["genie_annotation_pkg"] = genie_annotation_pkg

    # HACK: This is essential, because Synapse has concurrency update issues
    center_mapping_ent = syn.get(genie_config["centerMapping"])
    if center_mapping_ent.get("isProcessing", ["True"])[0] == "True":
        raise Exception(
            "Processing/validation is currently happening.  Please change/add the "
            f"'isProcessing' annotation on {genie_config['centerMapping']} "
            "to False to enable processing"
        )
    else:
        center_mapping_ent.isProcessing = "True"
        center_mapping_ent = syn.store(center_mapping_ent)

    # HACK: Create new maf database, should only happen once if its specified
    # Will modify genie configuration
    if create_new_maf_database:
        today = date.today()
        table_name = f"Narrow MAF Database - {today}"
        # filetype = "vcf2maf"
        # syn7208886 is the GENIE staging project to archive maf table
        new_tables = process_functions.create_new_fileformat_table(
            syn, "vcf2maf", table_name, project_id, "syn7208886"
        )
        syn.setPermissions(new_tables["newdb_ent"].id, 3326313, [])
        genie_config["vcf2maf"] = new_tables["newdb_ent"].id

    # Get file format classes
    format_registry = config.collect_format_types(args.format_registry_packages)

    # Start GENIE processing
    for process_center in centers:
        input_to_database.center_input_to_database(
            syn=syn,
            project_id=project_id,
            center=process_center,
            process=process,
            only_validate=only_validate,
            delete_old=delete_old,
            format_registry=format_registry,
            genie_config=genie_config,
        )

    # HACK: To ensure that this is the new entity
    center_mapping_ent = syn.get(genie_config["centerMapping"])
    center_mapping_ent.isProcessing = "False"
    # No need to return ent variable because it is unused
    syn.store(center_mapping_ent)

    error_tracker_synid = genie_config["errorTracker"]
    # Only write out invalid reasons if the center
    # isnt specified and if only validate
    if center is None and only_validate:
        logger.info("WRITING INVALID REASONS TO CENTER STAGING DIRS")
        write_invalid_reasons.write(
            syn, genie_config["centerMapping"], error_tracker_synid
        )
    logger.info("INPUT TO DATABASE COMPLETE")


if __name__ == "__main__":
    # Argument parsers
    # TODO: Fix case of arguments
    parser = argparse.ArgumentParser(description="GENIE center ")
    parser.add_argument(
        "process",
        choices=["mutation", "main"],
        help="Process vcf, maf or the rest of the files",
    )
    parser.add_argument(
        "--project_id", help="Synapse Project ID where data is stored.", required=True
    )
    parser.add_argument("--center", help="The centers")
    parser.add_argument("--pemFile", type=str, help="Path to PEM file (genie.pem)")
    parser.add_argument(
        "--deleteOld",
        action="store_true",
        help="Delete all old processed and temp files",
    )
    parser.add_argument(
        "--onlyValidate",
        action="store_true",
        help="Only validate the files, don't process",
    )
    parser.add_argument("--oncotree_link", type=str, help="Link to oncotree code")
    parser.add_argument(
        "--createNewMafDatabase", action="store_true", help="Creates a new maf database"
    )
    parser.add_argument(
        "--debug", action="store_true", help="Add debug mode to synapse"
    )
    parser.add_argument("--genie_annotation_pkg", help="GENIE annotation pkg")

    # DEFAULT PARAMS
    parser.add_argument(
        "--format_registry_packages",
        type=str,
        nargs="+",
        default=["genie_registry"],
        help="Python package name(s) to get valid file formats from "
        "(default: %(default)s).",
    )
    args = parser.parse_args()

    main(
        args.process,
        project_id=args.project_id,
        center=args.center,
        pemfile=args.pemFile,
        delete_old=args.deleteOld,
        only_validate=args.onlyValidate,
        oncotree_link=args.oncotree_link,
        create_new_maf_database=args.createNewMafDatabase,
        debug=args.debug,
        genie_annotation_pkg=args.genie_annotation_pkg,
        format_registry=args.format_registry_packages,
    )



from unittest import mock

import pandas as pd
import pytest
import synapseclient

from genie_registry.vcf import vcf

syn = mock.create_autospec(synapseclient.Synapse)

vcfClass = vcf(syn, "SAGE")


def test_processing():
    pass


@pytest.mark.parametrize("filepath_list", [["foo"], ["GENIE-SAGE-ID1.bed"]])
def test_incorrect_validatefilename(filepath_list):
    with pytest.raises(AssertionError):
        vcfClass.validateFilename(filepath_list)


def test_validation():
    assert vcfClass.validateFilename(["GENIE-SAGE-ID1.vcf"]) == "vcf"

    vcfDf = pd.DataFrame(
        {
            "#CHROM": ["2", "9", "12"],
            "POS": [69688533, 99401860, 53701241],
            "ID": ["AAK1", "AAED1", "AAAS"],
            "REF": ["AAK1", "AAED1", "AAAS"],
            "ALT": ["AAK1", "AAED1", "AAAS"],
            "QUAL": ["AAK1", "AAED1", "AAAS"],
            "FILTER": ["AAK1", "AAED1", "AAAS"],
            "INFO": ["AAK1", "AAED1", "AAAS"],
        }
    )

    error, warning = vcfClass._validate(vcfDf)
    assert error == ""
    assert warning == ""

    vcfDf = pd.DataFrame(
        {
            "POS": [69688533, 99401860, 53701241],
            "ID": ["AAK1", "AAED1", "AAAS"],
            "REF": ["AAK1", "AAED1", "AAAS"],
            "ALT": ["AAK1", "AAED1", "AAAS"],
            "QUAL": ["AAK1", "AAED1", "AAAS"],
            "FILTER": ["AAK1", "AAED1", "AAAS"],
            "INFO": ["AAK1", "AAED1", "AAAS"],
            "FOO": ["AAK1", "AAED1", "AAAS"],
            "DOO": ["AAK1", "AA ED1", "AAAS"],
        }
    )

    error, warning = vcfClass._validate(vcfDf)
    expectedError = (
        "vcf: Must have these headers: CHROM, POS, ID, REF, "
        "ALT, QUAL, FILTER, INFO.\n"
        "vcf: Must have FORMAT header if genotype columns exist.\n"
    )
    expectedWarning = (
        "vcf: Should not have any white spaces " "in any of the columns.\n"
    )
    assert error == expectedError
    assert warning == expectedWarning

    vcfDf = pd.DataFrame(
        {
            "#CHROM": ["chr2", "chrM", float("nan"), "chr2"],
            "POS": [69688533, 99401860, 53701241, 69688533],
            "ID": ["AAK1", "AAED1", "AAAS", "AAK1"],
            "REF": ["AAK1", "AAED1", "AAAS", "AAK1"],
            "ALT": ["AAK1", "AAED1", "AAAS", "AAK1"],
            "QUAL": ["AAK1", "AAED1", "AAAS", "AAK1"],
            "FILTER": ["AAK1", "AAED1", "AAAS", "AAK1"],
            "INFO": ["AAK1", "AAED1", "AAAS", "AAK1"],
        }
    )

    error, warning = vcfClass._validate(vcfDf)
    expectedError = (
        "vcf: Must not have duplicate variants.\n"
        "vcf: May contain rows that are "
        "space delimited instead of tab delimited.\n"
        "vcf: Must not have variants on chrM.\n"
    )
    expectedWarning = (
        "vcf: Should not have the chr prefix " "in front of chromosomes.\n"
    )
    assert error == expectedError
    assert warning == expectedWarning



from unittest import mock

import pandas as pd

from genie_registry.structural_variant import StructuralVariant


class TestSv:
    def setup_method(self, syn):
        sv_cls = StructuralVariant(syn, "SAGE")
        self.sv_cls = sv_cls

    def test_filetype(self):
        assert self.sv_cls._fileType == "sv"

    def test_processing(self):
        expected_df = pd.DataFrame(
            {
                "SAMPLE_ID": [
                    "GENIE-SAGE-ID1-1",
                    "GENIE-SAGE-ID2-1",
                    "GENIE-SAGE-ID3-1",
                    "GENIE-SAGE-ID4-1",
                    "GENIE-SAGE-ID5-1",
                ],
                "CENTER": ["SAGE", "SAGE", "SAGE", "SAGE", "SAGE"],
            }
        )
        sv_df = pd.DataFrame(
            {
                "sample_id": [
                    "GENIE-SAGE-ID1-1",
                    "GENIE-SAGE-ID2-1",
                    "GENIE-SAGE-ID3-1",
                    "GENIE-SAGE-ID4-1",
                    "GENIE-SAGE-ID5-1",
                ]
            }
        )
        processed_df = self.sv_cls._process(sv_df)
        assert expected_df.equals(processed_df)

    def test_validation_sample_error(self):
        sv_df = pd.DataFrame(
            {
                "sample_id": ["GENIE-SAGE-ID1-1", "GENIE-SAGE-ID1-1", "ID3-1"],
                "SV_STATUS": ["SOMATIC", "SOMATIC", "GERMLINE"],
                "SITE1_HUGO_SYMBOL": ["af", "af", "ff"],
                "SITE2_HUGO_SYMBOL": ["af", "af", "ff"],
            }
        )
        error, warning = self.sv_cls._validate(sv_df)
        assert error == (
            "Structural Variant: SAMPLE_ID must start with GENIE-SAGE\n"
            "Structural Variant: No duplicated rows allowed.\n"
        )
        assert warning == ""

    def test_validation_missing_required_cols(self):
        sv_df = pd.DataFrame(
            {
                "test": ["GENIE-SAGE-ID1-1", "GENIE-SAGE-ID1-2", "ID3-1"],
                "foo": ["SOMATIC", "SOMATIC", "GERMLINE"],
            }
        )
        error, warning = self.sv_cls._validate(sv_df)
        assert error == (
            "Structural Variant: Must have SAMPLE_ID column.\n"
            "Structural Variant: Must have SV_STATUS column.\n"
            "Structural Variant: Either SITE1_HUGO_SYMBOL/SITE1_ENTREZ_GENE_ID or "
            "SITE2_HUGO_SYMBOL/SITE2_ENTREZ_GENE_ID is required.\n"
        )
        assert warning == ""

    def test_validation_integer_check(self):
        sv_df = pd.DataFrame(
            {
                "sample_id": ["GENIE-SAGE-ID1-1", "GENIE-SAGE-ID2-1"],
                "SV_STATUS": ["SOMATIC", "GERMLINE"],
                "SITE1_ENTREZ_GENE_ID": [1, "foo"],
                "SITE2_ENTREZ_GENE_ID": [1, "foo"],
                "SITE1_REGION_NUMBER": [1, "foo"],
                "SITE2_REGION_NUMBER": [1, "foo"],
                "SITE1_POSITION": [1, "foo"],
                "SITE2_POSITION": [1, "foo"],
                "TUMOR_SPLIT_READ_COUNT": [1, "foo"],
                "TUMOR_PAIRED_END_READ_COUNT": [1, "foo"],
                "SV_LENGTH": [1, "foo"],
                "NORMAL_READ_COUNT": [1, "foo"],
                "TUMOR_READ_COUNT": [1, "foo"],
                "NORMAL_VARIANT_COUNT": [1, "foo"],
                "TUMOR_VARIANT_COUNT": [1, "foo"],
                "NORMAL_PAIRED_END_READ_COUNT": [1, "foo"],
                "NORMAL_SPLIT_READ_COUNT": [1, "foo"],
            }
        )
        error, warning = self.sv_cls._validate(sv_df)
        assert error == (
            "Structural Variant: Only integers allowed in these column(s): "
            "SITE1_ENTREZ_GENE_ID, SITE2_ENTREZ_GENE_ID, "
            "SITE1_REGION_NUMBER, SITE2_REGION_NUMBER, SITE1_POSITION, SITE2_POSITION, "
            "TUMOR_SPLIT_READ_COUNT, TUMOR_PAIRED_END_READ_COUNT, SV_LENGTH, "
            "NORMAL_READ_COUNT, TUMOR_READ_COUNT, NORMAL_VARIANT_COUNT, "
            "TUMOR_VARIANT_COUNT, NORMAL_PAIRED_END_READ_COUNT, "
            "NORMAL_SPLIT_READ_COUNT.\n"
        )
        assert warning == ""

    def test_validation_no_errors(self):
        sv_df = pd.DataFrame(
            {
                "sample_id": ["GENIE-SAGE-ID1-1", "GENIE-SAGE-ID2-1"],
                "SV_STATUS": ["SOMATIC", "GERMLINE"],
                "SITE1_ENTREZ_GENE_ID": [1, 2],
                "SITE2_ENTREZ_GENE_ID": [1, 3],
                "SITE1_REGION_NUMBER": [1, 2],
                "NCBI_BUILD": ["GRCh38", "GRCh37"],
                "BREAKPOINT_TYPE": ["PRECISE", "IMPRECISE"],
                "CONNECTION_TYPE": ["3to5", "5to5"],
                "DNA_SUPPORT": ["yes", "no"],
                "RNA_Support": ["yes", "no"],
            }
        )
        error, warning = self.sv_cls._validate(sv_df)
        assert error == ""
        assert warning == ""
